#!/usr/bin/env python
r"""
Screening computation with different parameters
===============================================

This examples shows how to compute the SCR file with different number of bands
and different values of ecuteps in order to prepare further GW convergence studies.
Each screening calculation is automatically parallelized over q-points and the
partial SCR files are then merged with the mrgscr utility.
The total SCR file is available in the outdata directory of the ScreeningWork.
"""
import sys
import os
import numpy as np

import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk


def build_flow(options):

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")


    # Initialize structure from string and a minimalistic variables:
    # natom, acell, rprimd and xred_symbols.

    structure = abilab.Structure.from_abistring("""
natom 2
acell  3*10.26         # Experimental lattice constants
rprim  0.0  0.5  0.5   # FCC primitive vectors (to be scaled by acell)
       0.5  0.0  0.5
       0.5  0.5  0.0

xred_symbols
      0.0  0.0  0.0  Si
      0.25 0.25 0.25 Si
""")

    # Build the SCF input.
    scf_inp = abilab.AbinitInput(structure=structure, pseudos=abidata.pseudos("14si.pspnc"))

    #nscf_nband_list = np.arange(2, 8, 2)
    nscf_nband_list = [25]
    max_nscf_nband = max(nscf_nband_list)

    #ecuteps_list = np.arange(2, 8, 2)
    ecuteps_list = [2]
    max_ecuteps = max(ecuteps_list)

    print("Number of valence electrons in the unit cell:", scf_inp.num_valence_electrons)

    scf_inp.set_vars(
        ecut=4,
        nband=8,
        ngkpt=[4, 4, 4],
        shiftk=[0, 0, 0],
        tolvrs=1e-8,
        timopt=-1,   # TODO: Add abirun.py option to extract timing data.
        #paral_kgb=1,
        #iomode=3,
    )

    #ebands_inp = scf_inp.make_ebands_input(ndivsm=15)

    # Build NSCF input on a k-mesh including empty states.
    nscf_inp = scf_inp.new_with_vars(
        nband=max_nscf_nband,
        #nbdbuf=10
        tolwfr=1e-20,   # Too high. Should be ~1e-20
        iscf=-2
    )

    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Band structure work to produce the WFK file with a NSCF run + empty states.
    bands_work = flowtk.BandStructureWork(scf_inp, nscf_inp) # dos_inputs=[nscf_inp])
    flow.register_work(bands_work)

    for ecuteps in ecuteps_list:
        for nscf_nband in nscf_nband_list:

            scr_inp = nscf_inp.new_with_vars(
                optdriver=3,
                ecuteps=ecuteps,
                nband=nscf_nband,
            )

            # Compute SCR file by splitting all the q-points.
            scr_work = flowtk.ScreeningWork.from_nscf_task(bands_work.nscf_task, scr_inp)

            # IMPORTANT:
            #   If you alredy have a WFK file and you want to skip the SCF + NSCF part
            #   build the scr_work using `from_wkf_filepath` instead of `from_nscf_task` e.g.:

            #scr_work = ScreeningWork.from_wfk_file(wfk_filepath, scr_inp)

            flow.register_work(scr_work)

            # This part is optional:
            #     1) set_readme sets the string that will be used to generate a human readable README.md file
            #     2) set_abipy_meta_json can be used to defined metadata that will be written to abipy_meta.json.
            scr_work.set_readme(f"SCR computation with ecuteps={ecuteps} and nband={nscf_nband}")
            scr_work.set_abipy_meta_json(dict(ecuteps=ecuteps, nband=nscf_nband))

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Relaxation + bands
==================

This script shows how to perform a structural relaxation
followed by a band structure calculation

WARNING: The API could change!
"""

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def make_ion_ioncell_inputs(paral_kgb=0):
    pseudos = abidata.pseudos("14si.pspnc")
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))

    # Perturb the structure (random perturbation of 0.1 Angstrom)
    # then compress the volume to trigger dilatmx.
    #structure.perturb(distance=0.1)
    #structure.scale_lattice(structure.volume * 0.6)

    global_vars = dict(
        ecut=28,
        ngkpt=[4,4,4],
        shiftk=[0,0,0],
        nshiftk=1,
        chksymbreak=0,
        paral_kgb=paral_kgb,
        #prtwf=0,
    )

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=2)

    # Global variables
    multi.set_vars(global_vars)

    # Dataset 1 (Atom Relaxation)
    multi[0].set_vars(
        optcell=0,
        ionmov=2,
        tolrff=0.02,
        tolmxf=5.0e-5,
        ntime=50,
        #ntime=3,  #To test the restart
        #dilatmx=1.1, # FIXME: abinit crashes if I don't use this
    )

    # Dataset 2 (Atom + Cell Relaxation)
    multi[1].set_vars(
        optcell=1,
        ionmov=2,
        ecutsm=0.5,
        dilatmx=1.02,
        tolrff=0.02,
        tolmxf=5.0e-5,
        strfact=100,
        ntime=50,
        #ntime=3,  # To test the restart
        )

    ion_inp, ioncell_inp = multi.split_datasets()
    return ion_inp, ioncell_inp


def make_scf_nscf_inputs(paral_kgb=1):
    """Returns two input files: GS run and NSCF on a high symmetry k-mesh."""
    pseudos = abidata.pseudos("14si.pspnc")
    #pseudos = data.pseudos("Si.GGA_PBE-JTH-paw.xml")

    multi = abilab.MultiDataset(structure=abidata.cif_file("si.cif"), pseudos=pseudos, ndtset=2)

    # Global variables
    ecut = 28
    global_vars = dict(ecut=ecut,
                       nband=8,
                       timopt=-1,
                       istwfk="*1",
                       nstep=15,
                       paral_kgb=paral_kgb,
                    )

    if multi.ispaw:
        global_vars.update(pawecutdg=2*ecut)

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(ngkpt=[8,8,8], shiftk=[0,0,0])
    multi[0].set_vars(tolvrs=1e-6)

    # Dataset 2 (NSCF run)
    kptbounds = [
        [0.5, 0.0, 0.0], # L point
        [0.0, 0.0, 0.0], # Gamma point
        [0.0, 0.5, 0.5], # X point
    ]

    multi[1].set_kpath(ndivsm=6, kptbounds=kptbounds)
    multi[1].set_vars(tolwfr=1e-12)

    # Generate two input files for the GS and the NSCF run
    scf_input, nscf_input = multi.split_datasets()
    return scf_input, nscf_input


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Create the flow
    flow = flowtk.Flow(options.workdir, manager=options.manager)

    paral_kgb = 1
    #paral_kgb = 0  # This one is OK

    # Create a relaxation work and add it to the flow.
    ion_inp, ioncell_inp = make_ion_ioncell_inputs(paral_kgb=paral_kgb)

    relax_work = flowtk.RelaxWork(ion_inp, ioncell_inp)
    flow.register_work(relax_work)

    scf_inp, nscf_inp = make_scf_nscf_inputs(paral_kgb=paral_kgb)

    bands_work = flowtk.BandStructureWork(scf_inp, nscf_inp)

    # The scf task in bands work restarts from the DEN file of the last task in relax_work
    if paral_kgb == 0:
        # cg works fine if we restart from the WFK
        bands_work.scf_task.add_deps({relax_work[-1]: "WFK"})
    else:
        # --> This triggers an infamous bug in abinit
        bands_work.scf_task.add_deps({relax_work[-1]: "WFK"})

        # --> This is ok if we used fourier_interp to change the FFT mesh.
        #bands_work.scf_task.add_deps({relax_work[-1]: "DEN"})

    # All task in bands_work will fetch the relaxed structure from the last task in relax_work
    for task in bands_work:
        task.add_deps({relax_work[-1]: "@structure"})

    flow.register_work(bands_work)
    flow.allocate()
    flow.use_smartio()
    flow.set_garbage_collector()

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Interface between phonopy and the AbiPy workflow model."""
from __future__ import annotations

import os
import numpy as np

try:
    from phonopy import Phonopy, file_IO
    from phonopy.interface.vasp import read_vasp_from_strings
    from phonopy.interface.abinit import parse_set_of_forces
except ImportError:
    import warnings
    warnings.warn("phonopy is required by abiphonopy. Install it with conda or pip install phonopy")

from abipy.core.structure import Structure
from abipy.flowtk.works import Work


__all__ = [
    "atoms_from_structure",
    "structure_from_atoms",
    "PhonopyWork",
    "PhonopyGruneisenWork",
]


def atoms_from_structure(structure: Structure):
    """
    Convert a pymatgen Structure into a phonopy Atoms object.
    """
    s = structure.to(fmt="poscar", filename=None)
    return read_vasp_from_strings(s, symbols=None)


def structure_from_atoms(atoms) -> Structure:
    """
    Convert a phonopy Atoms object into a abipy Structure.
    """
    return Structure(lattice=atoms.cell,
                     species=atoms.symbols,
                     coords=atoms.scaled_positions,
                     validate_proximity=False, to_unit_cell=False,
                     coords_are_cartesian=False, site_properties=None)


class PhonopyWork(Work):
    """
    This work computes the inter-atomic force constants with phonopy.

    .. attribute:: scdims(3)

        numpy arrays with the number of cells in the supercell along the three reduced directions

    .. attribute:: phonon

        :class:`Phonopy` object used to construct the supercells with displaced atoms.

    .. attribute:: phonopy_tasks

        List of :class:`ScfTask`. Each task compute the forces in one perturbed supercell.

    .. attribute:: bec_tasks

    .. attribute:: cpdata2dst

        If not None, the work will copy the output results to the outdir of the flow
        once all_ok is invoked. Note that cpdata2dst must be an absolute path.
    """

    @classmethod
    def from_gs_input(cls, gs_inp, scdims, phonopy_kwargs=None, displ_kwargs=None) -> PhonopyWork:
        """
        Build the work from an :class:`AbinitInput` object representing a GS calculations.

        Args:
            gs_inp::class:`AbinitInput` object representing a GS calculation in the initial unit cell.
            scdims: Number of unit cell replicas along the three reduced directions.
            phonopy_kwargs: (Optional) dictionary with arguments passed to Phonopy constructor.
            displ_kwargs: (Optional) dictionary with arguments passed to generate_displacements.

        Return: PhonopyWork instance.
        """
        new = cls()
        new.phonopy_tasks, new.bec_tasks = [], []
        new.cpdata2dst = None

        # Initialize phonon. Supercell matrix has (3, 3) shape.
        unitcell = atoms_from_structure(gs_inp.structure)
        new.scdims = scdims = np.array(scdims)
        if scdims.shape != (3,):
            raise ValueError("Expecting 3 int in scdims but got %s" % str(scdims))

        supercell_matrix = np.diag(scdims)
        phonopy_kwargs = phonopy_kwargs if phonopy_kwargs is not None else {}
        new.phonon = phonon = Phonopy(unitcell, supercell_matrix, **phonopy_kwargs)

        displ_kwargs = displ_kwargs if displ_kwargs is not None else {}
        phonon.generate_displacements(**displ_kwargs)  # distance=0.01,

        # Obtain supercells containing respective displacements (list of Atoms objects).
        for atoms in phonon.supercells_with_displacements:
            sc_struct = structure_from_atoms(atoms)
            sc_gsinp = gs_inp.new_with_structure(sc_struct, scdims=new.scdims)
            sc_gsinp.pop_tolerances()
            sc_gsinp.pop_vars(["ionmov", "optcell", "ntime"])
            sc_gsinp.set_vars(toldff=1.e-6)
            sc_gsinp.set_vars_ifnotin(chksymbreak=0, chkprim=0)

            task = new.register_scf_task(sc_gsinp)
            new.phonopy_tasks.append(task)

        return new

    def on_all_ok(self):
        """
        This method is called once the `Work` is completed i.e. when all the tasks
        have reached status S_OK. Here we get the forces from the output files, and
        we call phonopy to compute the inter-atomic force constants.
        """
        phonon = self.phonon

        # Write POSCAR with initial unit cell.
        structure = structure_from_atoms(phonon.get_primitive())
        structure.to(filename=self.outdir.path_in("POSCAR"))

        # Write yaml file with displacements.
        supercell = phonon.get_supercell()
        displacements = phonon.get_displacements()
        #directions = phonon.get_displacement_directions()
        file_IO.write_disp_yaml(displacements, supercell, # directions=directions,
                                filename=self.outdir.path_in('disp.yaml'))

        # Extract forces from the main Abinit output files.
        forces_filenames = [task.output_file.path for task in self.phonopy_tasks]
        num_atoms = supercell.get_number_of_atoms()
        force_sets = parse_set_of_forces(num_atoms, forces_filenames)

        # Write FORCE_SETS file.
        displacements = file_IO.parse_disp_yaml(filename=self.outdir.path_in('disp.yaml'))
        num_atoms = displacements['natom']
        for forces, disp in zip(force_sets, displacements['first_atoms']):
            disp['forces'] = forces
        file_IO.write_FORCE_SETS(displacements, filename=self.outdir.path_in('FORCE_SETS'))

        # Write README and configuration files.
        examples_url = "http://atztogo.github.io/phonopy/examples.html"
        doctags_url = "http://atztogo.github.io/phonopy/setting-tags.html#setting-tags"
        kptbounds = np.array([k.frac_coords for k in structure.hsym_kpoints])
        path_coords = " ".join(str(rc) for rc in kptbounds.flat)
        path_labels = " ".join(k.name for k in structure.hsym_kpoints)
        ngqpt = structure.calc_ngkpt(nksmall=30)

        with open(self.outdir.path_in("band.conf"), "wt") as fh:
            fh.write("#" + doctags_url + "\n")
            fh.write("DIM = %d %d %d\n" % tuple(self.scdims))
            fh.write("BAND = %s\n" % path_coords)
            fh.write("BAND_LABELS = %s\n" % path_labels)
            fh.write("BAND_POINTS = 101\n")
            fh.write("#BAND_CONNECTION = .TRUE.\n")

        with open(self.outdir.path_in("dos.conf"), "wt") as fh:
            fh.write("#" + doctags_url + "\n")
            fh.write("DIM = %d %d %d\n" % tuple(self.scdims))
            fh.write("MP = %d %d %d\n" % tuple(ngqpt))
            fh.write("#GAMMA_CENTER = .TRUE.\n")

        with open(self.outdir.path_in("band-dos.conf"), "wt") as fh:
            fh.write("#" + doctags_url + "\n")
            fh.write("DIM = %d %d %d\n" % tuple(self.scdims))
            fh.write("BAND = %s\n" % path_coords)
            fh.write("BAND_LABELS = %s\n" % path_labels)
            fh.write("BAND_POINTS = 101\n")
            fh.write("#BAND_CONNECTION = .TRUE.\n")
            fh.write("MP = %d %d %d\n" % tuple(ngqpt))
            fh.write("#GAMMA_CENTER = .TRUE.\n")

        with open(self.outdir.path_in("README.md"), "wt") as fh:
            fh.write("To plot bands, use:\n\tphonopy -p band.conf\n\n")
            fh.write("To plot phonon dos, use:\n\tphonopy -p dos.conf\n\n")
            fh.write("To plot bands and dos, use:\n\tphonopy -p band-dos.conf\n\n")
            fh.write("See also:\n")
            fh.write("\t" + examples_url + "\n")
            fh.write("\t" + doctags_url + "\n")

        if self.cpdata2dst:
            self.outdir.copy_r(self.cpdata2dst)

        return super().on_all_ok()


class PhonopyGruneisenWork(Work):
    """
    This work computes the Grüneisen parameters with phonopy. The workflow is as follows:

    It is necessary to run three phonon calculations.
    One is calculated at the equilibrium volume and the remaining two are calculated
    at the slightly larger volume and smaller volume than the equilibrium volume.
    The unitcells at these volumes have to be fully relaxed under the constraint of each volume.

    .. attribute:: scdims(3)

        numpy arrays with the number of cells in the supercell along the three reduced directions.
    """
    @classmethod
    def from_gs_input(cls, gs_inp, voldelta, scdims, phonopy_kwargs=None, displ_kwargs=None) -> PhonopyGruneisenWork:
        """
        Build the work from an :class:`AbinitInput` object representing a GS calculations.

        Args:
            gs_inp: :class:`AbinitInput` object representing a GS calculation in the initial unit cell.
            voldelta: Absolute increment for unit cell volume. The three volumes are:
                [v0 - voldelta, v0, v0 + voldelta] where v0 is taken from gs_inp.structure.
            scdims: Number of unit cell replicas along the three reduced directions.
            phonopy_kwargs: (Optional) dictionary with arguments passed to Phonopy constructor.
            displ_kwargs: (Optional) dictionary with arguments passed to generate_displacements.

        Return: `PhonopyGruneisenWork` instance.
        """
        new = cls()

        # Save arguments that will be used to call phonopy for creating
        # the supercells with the displacements once the three volumes have been relaxed.
        new.scdims = np.array(scdims)
        if new.scdims.shape != (3,):
            raise ValueError("Expecting 3 int in scdims but got %s" % str(new.scdims))
        new.phonopy_kwargs = phonopy_kwargs if phonopy_kwargs is not None else {}
        new.displ_kwargs = displ_kwargs if displ_kwargs is not None else {}

        # Build three tasks for structural optimization at constant volume.
        v0 = gs_inp.structure.volume
        if voldelta <= 0:
            raise ValueError("voldelta must be > 0 but got %s" % voldelta)
        volumes = [v0 - voldelta, v0, v0 + voldelta]
        if any(v <= 0 for v in volumes):
            raise ValueError("volumes must be > 0 but got %s" % str(volumes))

        for vol in volumes:
            # Build new structure
            new_lattice = gs_inp.structure.lattice.scale(vol)
            new_structure = Structure(new_lattice, gs_inp.structure.species, gs_inp.structure.frac_coords)
            new_input = gs_inp.new_with_structure(new_structure)
            # Set variables for structural optimization at constant volume.
            new_input.pop_tolerances()
            new_input.set_vars(optcell=3, ionmov=3, tolvrs=1e-10, toldff=1.e-6)
            new_input.set_vars_ifnotin(ecutsm=0.5, dilatmx=1.05)
            new.register_relax_task(new_input)

        return new

    def on_all_ok(self):
        """
        This method is called once the `Work` is completed i.e. when all the tasks
        have reached status S_OK.
        """
        self.add_phonopy_works_and_build()
        return super().on_all_ok()

    def add_phonopy_works_and_build(self):
        """
        Get relaxed structures from the tasks, build Phonopy works with supercells
        constructed from the new structures, add them to the flow and build new directories.
        """
        for i, task in enumerate(self):
            relaxed_structure = task.get_final_structure()
            gs_inp = task.input.new_with_structure(relaxed_structure)

            work = PhonopyWork.from_gs_input(gs_inp, self.scdims,
                                             phonopy_kwargs=self.phonopy_kwargs,
                                             displ_kwargs=self.displ_kwargs)

            self.flow.register_work(work)
            # Tell the work to copy the results to e.g. `flow/outdir/w0/minus`
            dst = os.path.join(self.pos_str, {0: "minus", 1: "orig", 2: "plus"}[i])
            work.cpdata2dst = self.flow.outdir.path_in(dst)

        # Allocate new works and update the pickle database.
        self.flow.allocate()
        self.flow.build_and_pickle_dump()



# coding: utf-8
"""
This modules provides subclasses of pymatgen tensor objects.
"""
from __future__ import annotations
import numpy as np
import pandas as pd

from pymatgen.core.tensors import Tensor, SquareTensor
from pymatgen.analysis.elasticity.elastic import ElasticTensor  # noqa: F401
from pymatgen.analysis.elasticity.stress import Stress as pmg_Stress
from pymatgen.analysis.piezo import PiezoTensor # noqa: F401
from abipy.iotools import ETSF_Reader


class _Tensor33:

    def _repr_html_(self):
        """Integration with jupyter notebooks."""
        return self.get_dataframe()._repr_html_()

    def get_dataframe(self, tol=1e-3, cmode=None) -> pd.DataFrame:
        """
        Return |pandas-Dataframe| with tensor elements set to zero below `tol`.

        Args:
            cmode: "real" or "imag" to include only the real/imaginary part.
        """
        tensor = self.zeroed(tol=tol)
        if cmode == "real": tensor = tensor.real
        if cmode == "imag": tensor = tensor.imag

        return pd.DataFrame({"x": tensor[:,0], "y": tensor[:,1], "z": tensor[:,2]}, index=["x", "y", "z"])

    def get_voigt_dataframe(self, tol=1e-3) -> pd.DataFrame:
        """
        Return |pandas-DataFrame| with Voigt indices as colums.
        Elements below tol are set to zero.

        Useful to analyze the converge of individual elements.
        """
        tensor = self.zeroed(tol=tol)
        columns = ["xx", "yy", "zz", "yz", "xz", "xy"]
        d = {k: v for k, v in zip(columns, tensor.voigt)}
        return pd.DataFrame(d, index=[0], columns=columns)


class Stress(pmg_Stress, _Tensor33):
    """
    Stress tensor. rank2 symmetric tensor with shape [3, 3].

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: Stress
    """


class DielectricTensor(SquareTensor, _Tensor33):
    """
    Subclass of |pmg-Tensor| describing a dielectric tensor.
    rank2 symmetric tensor with shape [3, 3].

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DielectricTensor
    """

    def reflectivity(self, n1=1, tol=1e-6) -> pd.DataFrame:
        """
        If the tensor is diagonal (with off diagonal elements smaller than tol)
        returns the three components of the reflectivity

            :math:`|n1 - n2| / | n1 + n2 |`
        """
        d = np.diag(self)

        if np.max(np.abs(self - np.diag(d))) > tol:
            raise ValueError("The tensor is not diagonal.")

        n2 = np.sqrt(d)

        return np.abs((n1 - n2) / (n1 + n2)) ** 2


class ZstarTensor(SquareTensor, _Tensor33):
    """
    Born effective charge tensor (for a single atom).

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: ZstarTensor
    """


class NLOpticalSusceptibilityTensor(Tensor):
    """
    Subclass of |pmg-Tensor| containing the non-linear optical susceptibility tensor.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: NLOpticalSusceptibilityTensor
    """

    @classmethod
    def from_file(cls, filepath: str) -> NLOpticalSusceptibilityTensor:
        """
        Creates the tensor from an anaddb.nc netcdf file containing ``dchide``.
        This requires to run anaddb with ``tnlflag`` > 0
        """
        with ETSF_Reader(filepath) as reader:
            try:
                return cls(reader.read_value("dchide"))
            except Exception as exc:
                import traceback
                msg = traceback.format_exc()
                msg += ("Error while trying to read from file.\n"
                        "Verify that nlflag > 0 in anaddb\n")
                raise ValueError(msg)



"""
"""
from __future__ import annotations

import os
import warnings
import numpy as np

from fnmatch import fnmatch
from monty.string import list_strings
from abipy.core.structure import Structure
from abipy.dynamics.hist import HistFile


def get_energy_step(step: dict) -> float:
    """
    Copied from final_energy property in vasp.outputs.

    Addresses a bug in vasprun.xml. See https://www.vasp.at/forum/viewtopic.php?f=3&t=16942
    """
    final_istep = step
    total_energy = final_istep["e_0_energy"]
    final_estep = final_istep["electronic_steps"][-1]
    electronic_energy_diff = final_estep["e_0_energy"] - final_estep["e_fr_energy"]
    total_energy_bugfix = np.round(electronic_energy_diff + final_istep["e_fr_energy"], 8)
    if np.abs(total_energy - total_energy_bugfix) > 1e-7:
        return total_energy_bugfix
    return total_energy


def get_structures_labels_from_file(filepath: str) -> tuple[list[Structure], dict]:
    """
    Read energies, forces, stresses and magmoms from an external file. 
    Return list of structures and dict with results.
    """
    basename = os.path.basename(str(filepath))

    if basename.endswith("_HIST.nc"):
        with HistFile(filepath) as hist:
            structures = hist.reader.read_all_structures()
            # etotals per atom in eV.
            energies_per_atom = [float(e) / hist.r.natom for e in hist.etotals]
            cart_forces = hist.r.read_cart_forces(unit="eV ang^-1")
            # GPa units.
            stress_cart_tensors, pressures = hist.r.read_cart_stress_tensors()
            labels = {
                "energies": energies_per_atom,
                "forces": cart_forces,
                "stresses": stress_cart_tensors,
                #"magmoms": None,
            }

    elif fnmatch(basename, "vasprun*.xml*"):
        from pymatgen.io.vasp.outputs import Vasprun
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            vasprun = Vasprun(filepath)

        structures = vasprun.structures
        natom = len(structures[0])
        energies_per_atom, forces, stresses, magmoms = [], [], [], []
        for step in vasprun.ionic_steps:
            energies_per_atom.append(get_energy_step(step) / natom)
            forces.append(step["forces"])
            stresses.append(step["stress"])

        labels = {
            "energies": energies_per_atom,
            "forces": forces,
            "stresses": stresses,
            #"magmoms": None,
        }

    else:
        raise ValueError(f"Don't know how to extract data from: {filepath=}")

    return structures, labels


def get_structures_labels_from_files(filepaths) -> tuple[list[Structure], dict]:
    """
    Read energies, forces, stresses and magmoms from an external file. 
    Return list of structures and dict with results.
    """
    for i, path in enumerate(list_strings(filepaths)):
        this_structures, this_labels = get_structures_labels_from_file(path)
        if i == 0:
            structures, labels = this_structures, this_labels
        else:
            structures += this_structures
            for k in labels:
                if labels[k] is None: continue
                labels[k] += this_labels[k]

    #for s in structures: print(s)
    #for k, v in labels.items(): print(k, v)

    return structures, labels




# coding: utf-8
# flake8: noqa
from __future__ import annotations

import numpy as np
from pymatgen.io.abinit.netcdf import as_etsfreader
try:
    from pymatgen.io.abinit.netcdf import ETSF_Reader
except ImportError:
    from pymatgen.io.abinit.netcdf import EtsfReader as ETSF_Reader

from monty.functools import lazy_property
from pymatgen.core.periodic_table import Element
from .xsf import *
from .visualizer import *


#as_etsfreader = ionc.as_etsfreader


class ETSF_Reader(ETSF_Reader):
    """
    Provides high-level API to read data from netcdf files written
    following the ETSF-IO specifications described in :cite:`Caliste2008`
    """

    def read_structure(self):
        """
        Overrides the ``read_structure`` method so that we always return
        an instance of AbiPy |Structure| object
        """
        from abipy.core.structure import Structure
        return Structure.from_file(self.path)

    def typeidx_from_symbol(self, symbol: str) -> int:
        """Returns the type index from the chemical symbol. Note python convention."""
        return self.chemical_symbols.index(symbol)

    # Must overwrite implementation of pymatgen.io.abinit.netcdf
    # due to a possible bug introduced by initial whitespaces in symbol
    @lazy_property
    def chemical_symbols(self):
        """Chemical symbols char [number of atom species][symbol length]."""
        charr = self.read_value("chemical_symbols")
        symbols = []
        for v in charr:
            s = "".join(c.decode("utf-8") for c in v)
            # Strip to avoid possible whitespaces.
            symbols.append(s.strip())

        return symbols

    def read_string(self, varname: str):
        """
        Args:
            varname: Name of the variable
        """
        b = self.rootgrp.variables[varname][:]
        #print(type(b))
        import netCDF4
        try:
            value = netCDF4.chartostring(b)[()].decode('utf-8')
        except Exception:
            try:
                value = netCDF4.chartostring(b)[()]
            except Exception:
                try:
                    value = "".join(c for c in self.read_value(varname))
                except TypeError as exc:
                    value = "".join(c.decode("utf-8") for c in self.read_value(varname))

        return value.strip()

    def none_if_masked_array(self, arr):
        """Return None if arr is a MaskedArray else None."""
        return None if np.ma.is_masked(arr) else arr

    def read_amu_symbol(self) -> dict:
        """
        Read atomic masses and return dictionary element_symbol --> amu.

        .. note::

            Only netcdf files with phonon-related quantities contain this variable.
        """
        for k in ("atomic_mass_units", "atomic_numbers"):
            if k not in self.rootgrp.variables:
                raise RuntimeError("`%s` does not contain `%s` variable." % (self.path, k))

        # ntypat arrays
        amu_list = self.read_value("atomic_mass_units")
        atomic_numbers = self.read_value("atomic_numbers")
        amu_z = {at: a for at, a in zip(atomic_numbers, amu_list)}
        amu_symbol = {Element.from_Z(n).symbol: v for n, v in amu_z.items()}

        return amu_symbol

    def read_ngfft3(self) -> np.array:
        """
        Return the number of FFT divisions.
        """
        ngfft3 = 3 * [None]
        ngfft3[0] = self.read_dimvalue("number_of_grid_points_vector1")
        ngfft3[1] = self.read_dimvalue("number_of_grid_points_vector2")
        ngfft3[2] = self.read_dimvalue("number_of_grid_points_vector3")
        return np.array(ngfft3, int)



#!/usr/bin/env python
r"""
Quasi-harmonic approximation (convergence wrt Q-mesh)
=====================================================

This example shows how to use the GSR.nc and PHDOS.nc files computed with different volumes
to analyze the convergence of QHA thermodynamic properties with respect to the number of q-points
used to compute the phonon DOS.
"""
import os
import abipy.data as abidata

from abipy.dfpt.qha import QHAQmeshAnalyzer

# We use a list of GSR.nc and PHDOS.nc files corresponding to different isotropic strains.
# These files are shipped with AbiPy so that we don't need to run calculations from scratch.
#strains = [-4, -2, 0, 2, 4, 6]
strains = [-2, 0, 2, 4]
dirpath = os.path.join(abidata.dirpath, "refs", "si_qha")

gsr_paths = [os.path.join(dirpath, "mp-149_{:+d}_GSR.nc".format(s)) for s in strains]
ddb_paths = [os.path.join(dirpath, "mp-149_{:+d}_DDB".format(s)) for s in strains]

# Initialize QHA object from files.
qhana = QHAQmeshAnalyzer(gsr_paths, ddb_paths)

# To change the default EOS (vinet), use
#qhana.set_eos("vinet")

# Compute ph-DOS with this list of nqsmall values.
qhana.run_qlist([2, 4])

# Analyze convergence. API similar to QHA.
qhana.plot_energies(title="Energies as a function of volume for different T")

qhana.plot_thermal_expansion_coeff(title="Thermal expansion coefficient as a function of T")

qhana.plot_vol_vs_t(title="Volume as a function of T")



# coding: utf-8
"""Tools to analyze the output files produced by Lobster."""
from __future__ import annotations

import os
import re
import glob
import numpy as np

from collections import defaultdict, OrderedDict
from monty.string import marquee
from monty.collections import tree
from monty.io import zopen
from monty.termcolor import cprint
from monty.functools import lazy_property
from pymatgen.core.periodic_table import Element
from pymatgen.electronic_structure.core import OrbitalType
from pymatgen.io.abinit.pseudos import Pseudo
from abipy.core.func1d import Function1D
from abipy.core.mixins import BaseFile, NotebookWriter
from abipy.electrons.gsr import GsrFile
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_visible, set_ax_xylabels
from abipy.tools.typing import Figure
from abipy.tools import duck


class _LobsterFile(BaseFile, NotebookWriter):
    """
    Base class for output files produced by lobster.
    """
    # These class attributes can be redefined in order to customize the plots.

    # Mapping L --> color used in plots.
    color_l = {"s": "red", "p": "blue", "d": "green", "f": "orange"}

    # \U starts an eight-character Unicode escape. raw strings do not work in python2.7
    # and we need a latex symbol to avoid errors in matplotlib --> replace myuparrow --> uparrow

    # Mapping spin --> title used in subplots that depend on (collinear) spin.
    spin2tex = {k: v.replace("myuparrow", "uparrow") for k, v in
            {0: r"$\sigma=\myuparrow$", 1: r"$\sigma=\downarrow$"}.items()}

    # TODOL ls
    params_orbs = {
      "s":  {"style": dict(color="red", ls="-", lw=1), "latex": "s"},
      #
      "p_x": {"style": dict(color="blue", ls="-", lw=1), "latex": "p_x"},
      "p_y": {"style": dict(color="blue", ls="-", lw=1), "latex": "p_y"},
      "p_z": {"style": dict(color="blue", ls="-", lw=1), "latex": "p_z"},
      #
      "d_xy":  {"style": dict(color="green", ls="-", lw=1), "latex": "d_{xy}"},
      "d_yz":  {"style": dict(color="green", ls="-", lw=1), "latex": "d_{yz}"},
      "d_z^2": {"style": dict(color="green", ls="-", lw=1), "latex": "d_{z^2}"},
      "d_xz":  {"style": dict(color="green", ls="-", lw=1), "latex": "d_{xz}"},
      "d_x^2-y^2": {"style": dict(color="green", ls="-", lw=1), "latex": "d_{x^2-y^2}"},
      #
      "4f_y(3x^2-y^2)": {"style": dict(color="orange", ls="-", lw=1), "latex": "4f_{y(3x^2-y^2)}"},
      "4f_xyz": {"style": dict(color="orange", ls="-", lw=1), "latex": "4f_{xyz}"},
      "4f_yz^2": {"style": dict(color="orange", ls="-", lw=1), "latex": "4f_{yz^2}"},
      "4f_z^3": {"style": dict(color="orange", ls="-", lw=1), "latex": "4f_{z^3}"},
      "4f_xz^2": {"style": dict(color="orange", ls="-", lw=1), "latex": "4f_{xz^2}"},
      "4f_z(x^2-y^2)": {"style": dict(color="orange", ls="-", lw=1), "latex": "4f_{z(x^2-y^2)}"},
    }

    def __str__(self) -> str:
        return self.to_string()

    def close(self) -> None:
        """Needed by ABC."""

    #@add_fig_kwargs
    #def plot_with_ebands(self, ebands, fontsize=12, **kwargs) -> Figure:
    #    """
    #    Plot bands + (COHP|COOP|DOSCAR) depending on the content of the file.

    #    Args:
    #        ebands: Path to ncfile with ebands or |ElectronBands| object.
    #        fontsize: fontsize for legends and titles

    #    Returns: |matplotlib-Figure|
    #    """
    #    ebands = ElectronBands.as_ebands(ebands)

    #    import matplotlib.pyplot as plt
    #    from matplotlib.gridspec import GridSpec
    #    fig = plt.figure()

    #    # Build grid.
    #    nrows, ncols = 1, 2
    #    gspec = GridSpec(nrows=nrows, ncols=ncols, width_ratios=(2, 1), wspace=0.05)

    #    # Bands and DOS will share the y-axis
    #    axmat = np.array((nrows, ncols), dtype=object)
    #    for icol in range(ncols):
    #        for irow in range(nrows):
    #            axmat[irow, icol] = plt.subplot(gspec[irow, icol], sharey=None if irow == 0 else axmat[irow, icol-1])

    #    #axmat, fig, plt = get_axarray_fig_plt(None, nrows=self.nsppol, ncols=len(entries) + 1,
    #    #                                        sharex=False, sharey=True, squeeze=False)

    #    for ix, ax in enumerate(axmat.ravel()):
    #        if ix == 0:
    #            ebands.plot(e0="fermie", ax=ax, show=False)
    #        else:
    #            self.plot(ax=ax, exchange_xy=True, show=False)

    #        #if ix != 0:
    #        #    set_visible(ax, False, "ylabel")
    #        #if self.nsppol == 2 and spin == 0:
    #        #    set_visible(ax, False, "xlabel")

    #    return fig


class CoxpFile(_LobsterFile):
    """
    Wrapper class for the crystal orbital projections produced by Lobster.
    Wraps both a COOP and a COHP.
    Can contain both the total and orbitalwise projections.

    .. attribute:: cop_type

        String. Either "coop" or "cohp".

    .. attribute:: type_of_index

        Dictionary mappping site index to element string.

    .. attribute:: energies

        List of energies. Shifted such that the Fermi level lies at 0 eV.

    .. attribute:: total

        A dictionary with the values of the total overlap, not projected over orbitals.
        The dictionary should have the following nested keys: a tuple with the index of the sites
        considered as a pair (0-based, e.g. (0, 1)), the spin (i.e. 0 or 1), a "single"
        or "integrated" string indicating the value corresponding to the value of
        the energy or to the integrated value up to that energy.

    .. attribute:: partial

        A dictionary with the values of the partial crystal orbital projections.
        The dictionary should have the following nested keys: a tuple with the index of the sites
        considered as a pair (0-based, e.g. (0, 1)), a tuple with the string representing the
        projected orbitals for that pair (e.g. ("4s", "4p_x")), the spin (i.e. 0 or 1),
        a "single" or "integrated" string indicating the value corresponding to the value of
        the energy or to the integrated value up to that energy. Each dictionary should contain a
        numpy array with a list of COP values with the same size as energies.

    .. attribute:: averaged

        A dictionary with the values of the partial crystal orbital projections
        averaged over all atom pairs specified. The main key should indicate the spin (0 or 1)
        and the nested dictionary should have "single" and "integrated" as keys.

    .. attribute:: fermie

        value of the fermi energy in eV.
    """

    @property
    def site_pairs_total(self) -> list:
        """
        List of site pairs available for the total COP
        """
        return list(self.total.keys())

    @property
    def site_pairs_partial(self) -> list:
        """
        List of site pairs available for the partial COP
        """
        return list(self.partial.keys())

    @classmethod
    def from_file(cls, filepath: str) -> CoxpFile:
        """
        Generates an instance of CoxpFile from the files produce by Lobster.
        Accepts gzipped files.

        Args:
            filepath: path to the COHPCAR.lobster or COOPCAR.lobster.

        Returns:
            A CoxpFile.
        """
        # From lobster documentation
        # COHPCAR.lobster:
        # File that contains the pCOHPs as requested in the lobsterin file.
        # It resembles the format of TB-LMTO-ASA's COPL file, which is organized as follows:
        # - Starting in line 3, the labels for the interactions are presented, followed by the
        # actual data.
        # - Column 1: energy axis, shifted such that the Fermi level lies at zero eV.
        # - Column 2: pCOHP averaged over all atom pairs specified
        # - Column 3: integrated pCOHP (IpCOHP) averaged over all atom pairs
        # - Column 4: pCOHP of the first interaction
        # - Column 5: IpCOHP of the first interaction
        # - and so on...
        # Note that in a spin-polarized calculation, the first set of the columns (2, 3, ..., 2N+3)
        # belongs to the first (up) spin and the other set (2N+4, 2N+5, ..., 4N+5) belongs to the
        # second (down) spin. Here N is the number of interactions.

        float_patt = r'-?(?:0|[1-9]\d*)(?:\.\d*)?(?:[eE][+\-]?\d+)?'
        header_patt = re.compile(r'\s+(\d+)\s+(\d+)\s+(\d+)\s+(' + float_patt +
                                 r')\s+(' + float_patt + r')\s+(' + float_patt + r')')
        pair_patt = re.compile(r'No\.\d+:([a-zA-Z]+)(\d+)(?:\[([a-z0-9_\-^]+)\])?->([a-zA-Z]+)(\d+)(?:\[([a-z0-9_\-^]+)\])?')

        new = cls(filepath)

        with zopen(filepath, "rt") as f:
            # Find the header
            for line in f:
                match = header_patt.match(line.rstrip())
                if match:
                    n_column_groups = int(match.group(1))
                    n_spin = int(match.group(2))
                    n_en_steps = int(match.group(3))
                    min_en = float(match.group(4))
                    max_en = float(match.group(5))
                    new.fermie = float(match.group(6))
                    break
            else:
                raise ValueError("Can't find the header in file {}".format(filepath))

            n_pairs = n_column_groups-1

            count_pairs = 0
            pairs_data = []
            new.type_of_index = {}
            # Parse the pairs considered
            for line in f:
                match = pair_patt.match(line.rstrip())
                if match:
                    # Adds a tuple: [type1, index1, orbital1, type2, index2, orbital2]
                    # with orbital1, orbital2 = None if the pair is not orbitalwise
                    type1, index1, orbital1, type2, index2, orbital2 = match.groups()
                    # 0-based indexing
                    index1 = int(index1) - 1
                    index2 = int(index2) - 1
                    pairs_data.append([type1, index1, orbital1, type2, index2, orbital2])
                    if index1 in new.type_of_index: assert new.type_of_index[index1] == type1
                    new.type_of_index[index1] = type1
                    if index2 in new.type_of_index: assert new.type_of_index[index2] == type2
                    new.type_of_index[index2] = type2
                    count_pairs += 1
                    if count_pairs == n_pairs:
                        break

            spins = [0, 1][:n_spin]

            data = np.fromstring(f.read(), dtype=float, sep=' ').reshape([n_en_steps, 1+n_spin*n_column_groups*2])

            # Initialize and fill results
            new.energies = data[:, 0]
            new.averaged = defaultdict(dict)
            new.total = tree()
            new.partial = tree()

            for i, s in enumerate(spins):
                base_index = 1+i*n_column_groups*2
                new.averaged[s]['single'] = data[:, base_index].copy()
                new.averaged[s]['integrated'] = data[:, base_index+1].copy()
                # NB (i, j) --> (j, i) symmetry is enforced to make API easier.
                for j, p in enumerate(pairs_data):
                    index1 = p[1]
                    index2 = p[4]
                    if p[2] is not None:
                        # Partial
                        single = data[:, base_index+2*(j+1)].copy()
                        integrated = data[:, base_index+2*(j+1)+1].copy()
                        new.partial[(index1, index2)][(p[2], p[5])][s]['single'] = single
                        new.partial[(index2, index1)][(p[5], p[2])][s]['single'] = single
                        new.partial[(index1, index2)][(p[2], p[5])][s]['integrated'] = integrated
                        new.partial[(index2, index1)][(p[5], p[2])][s]['integrated'] = integrated
                    else:
                        # Total
                        single = data[:, base_index+2*(j+1)].copy()
                        integrated = data[:, base_index+2*(j+1)+1].copy()
                        new.total[(index1, index2)][s]['single'] = single
                        new.total[(index2, index1)][s]['single'] = single
                        new.total[(index1, index2)][s]['integrated'] = integrated
                        new.total[(index2, index1)][s]['integrated'] = integrated

        new.cop_type = "unknown"
        if "COOPCAR.lobster" in filepath: new.cop_type = "coop"
        if "COHPCAR.lobster" in filepath: new.cop_type = "cohp"
        new.nsppol = len(new.averaged)

        return new

    @lazy_property
    def functions_pair_lorbitals(self):
        """
        Extracts a dictionary with keys pair, orbital, spin and containing a |Function1D| object resolved
        for l orbitals.
        """
        if not self.partial:
            raise RuntimeError("Partial orbitals not calculated.")

        results = tree()
        for pair, pair_data in self.partial.items():
            # Check if the symmetric has already been calculated
            if (pair[1], pair[0]) in results:
                for orbs, orbs_data in results[(pair[1], pair[0])].items():
                    results[pair][(orbs[1], orbs[0])] = orbs_data
                continue

            # For each look at all orbital possibilities
            for orbs, orbs_data in pair_data.items():
                k = (orbs[0].split("_")[0], orbs[1].split("_")[0])
                if k in results[pair]:
                    for spin in orbs_data.keys():
                        results[pair][k][spin] = results[pair][k][spin] + Function1D(self.energies,orbs_data[spin]['single'])
                else:
                    for spin in orbs_data.keys():
                        results[pair][k][spin] = Function1D(self.energies, orbs_data[spin]['single'])

        return results

    @lazy_property
    def functions_pair_morbitals(self):
        """
        Extracts a dictionary with keys pair, orbital, spin and containing a |Function1D| object resolved
        for l and m orbitals.
        """
        if not self.partial:
            raise RuntimeError("Partial orbitals not calculated.")

        results = tree()
        for pair, pair_data in self.partial.items():
            for orbs, orbs_data in pair_data.items():
                for spin in orbs_data.keys():
                    results[pair][orbs][spin] = Function1D(self.energies, orbs_data[spin]['single'])
        return results

    @lazy_property
    def functions_pair(self):
        """
        Extracts a dictionary with keys pair, spin and containing a |Function1D| object for the total COP.
        """
        results = tree()
        for pair, pair_data in self.total.items():
            for spin in pair_data.keys():
                results[pair][spin] = Function1D(self.energies, pair_data[spin]['single'])
        return results

    def to_string(self, verbose=0) -> str:
        """String representation with verbosity level `verbose`."""
        lines = []; app = lines.append
        if verbose:
            #app(marquee("File Info", mark="="))
            app(self.filestat(as_string=True))
        app("")
        app("%s: Number of energies: %d, from %.3f to %.3f (eV) with E_fermi set 0 (was %.3f)" % (
            self.cop_type.upper(), len(self.energies), self.energies[0], self.energies[-1], self.fermie))
        app("has_partial_projections: %s, nsppol: %d" % (bool(self.partial), self.nsppol))
        app("Number of pairs: %d" % len(self.total))
        for i, pair in enumerate(self.total):
            type0, type1 = self.type_of_index[pair[0]], self.type_of_index[pair[1]]
            app("[%d] %s@%s --> %s@%s" % (i, type0, pair[0], type1, pair[1]))

        return "\n".join(lines)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(what="d", show=False)
        yield self.plot(what="i", show=False)

    @add_fig_kwargs
    def plot(self, what="d", spin=None, ax=None, exchange_xy=False, fontsize=12, **kwargs) -> Figure:
        """
        Plot COXP averaged values (DOS or IDOS depending on what).

        Args:
            what: string selecting what will be plotted. "d" for DOS, "i" for IDOS
            spin: Select spin index if not None.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            exchange_xy: True to exchange x-y axis.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        if not self.averaged: return None
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)

        xlabel = r"$E - E_f\;(eV)$"
        if self.cop_type == "coop":
            ysign = +1
            ylabel = {"d": "COOP", "i": "ICOOP"}[what]
        elif self.cop_type == "cohp":
            ysign = -1
            ylabel = {"d": "-COHP", "i": "-ICOHP"}[what]
        else:
            raise ValueError("Wrong cop_type: `%s`" % str(self.cop_type))

        spins = range(self.nsppol) if spin is None else [spin]
        for spin in spins:
            opts = {"color": "black", "linewidth": 2.0, "ls": "-"} if spin == 0 else \
                   {"color": "red", "linewidth": 2.0, "ls": "--"}
            opts.update(kwargs)
            key = {"d": "single", "i": "integrated"}[what]
            xs, ys = self.energies, ysign * self.averaged[spin][key]
            if exchange_xy: xs, ys = ys, xs
            ax.plot(xs, ys, label=self.spin2tex[spin] if self.nsppol == 2 else None,
                    **opts)

        if exchange_xy:
            # Add vertical line to signal the zero.
            ax.axvline(c="k", ls=":", lw=1)
        else:
            ax.axhline(c="k", ls=":", lw=1)

        set_ax_xylabels(ax, xlabel, ylabel, exchange_xy)

        if self.nsppol == 2:
            ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_average_pairs(self, with_site_index, what="single", exchange_xy=False,
                           fontsize=8, **kwargs) -> Figure:
        """
        Plot COXP total overlap for all sites containg `with_site_index` and average sum
        (multiplied by the number of pairs)

        Args:
            with_site_index: int of list of integers selecting the site index to be included.
            what: "single" for COXP DOS, "integrated" for IDOS.
            exchange_xy: True to exchange x-y axis.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        if not duck.is_listlike(with_site_index): with_site_index = [with_site_index]

        # Create list of pairs.
        all_pairs = []
        for index in with_site_index:
            pairs = [p for p in self.total if index in p]
            all_pairs.append(pairs)
            if not pairs:
                cprint("Cannot find pairs containing  site index %s" % str(index), "yellow")
            else:
                print("Pairs containing site index:", index)
                for pair in pairs:
                    type0, type1 = self.type_of_index[pair[0]], self.type_of_index[pair[1]]
                    print("\t%s@%s --> %s@%s" % (type0, pair[0], type1, pair[1]))

        # Build (1, nsppol) grid
        #ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=1, ncols=self.nsppol,
                                                sharex=True, sharey=True, squeeze=False)

        xlabel, ylabel, ysign = r"$E - E_f\;(eV)$", "COOP", +1
        if self.cop_type == "cohp":
            ylabel, ysign = "-COHP", -1

        # self.total[pair][spin][what]
        for spin, ax in enumerate(ax_list.ravel()):
            ax.grid(True)
            sum_all = np.zeros_like(self.energies)
            for index, pairs in zip(with_site_index, all_pairs):
                if not pairs: continue
                ys = np.zeros_like(self.energies)
                for pair in pairs:
                    ys += self.total[pair][spin][what]
                #ys /= len(pairs)
                ys /= len(self.total)
                sum_all += ys

                #label, style = self.get_labelstyle_from_spin_pair(spin, pair)
                label = "Average over pairs with %s@%s" % (self.type_of_index[index], index)
                style = {}
                xs, ys = self.energies, ysign * ys
                if exchange_xy: xs, ys = ys, xs
                ax.plot(xs, ys, label=label, **style)

            # Plot average * num_pairs for this spin.
            xs, ys = self.energies, ysign * self.averaged[spin][what] * 2 * len(self.total)
            if exchange_xy: xs, ys = ys, xs
            style = {}
            ax.plot(xs, ys, label="Average", **style)

            # Compare with sum
            xs, ys = self.energies, ysign * sum_all
            if exchange_xy: xs, ys = ys, xs
            ax.plot(xs, ys, label="Sumall", **style)

            if exchange_xy:
                # Add vertical line to signal the zero.
                ax.axvline(c="k", ls=":", lw=1)
            else:
                ax.axhline(c="k", ls=":", lw=1)

            if self.nsppol == 2: ax.set_title(self.spin2tex[spin])
            set_ax_xylabels(ax, xlabel, ylabel, exchange_xy)
            ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_site_pairs_total(self, from_site_index, what="single", exchange_xy=False, ax=None,
                              fontsize=8, **kwargs) -> Figure:
        """
        Plot COXP total overlap (DOS or IDOS) for all sites listed in `from_site_index`

        Args:
            from_site_index: int of list of integers selecting the first site of the pairs to be included.
            what: "single" for COXP DOS, "integrated" for IDOS.
            exchange_xy: True to exchange x-y axis.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        # Handle list of sites.
        if duck.is_listlike(from_site_index):
            nrows = len(from_site_index)
            sharex, sharey = True, False
            if exchange_xy: sharex, sharey = sharey, sharex
            ax_list, fig, plt = get_axarray_fig_plt(ax, nrows=nrows, ncols=1,
                                                    sharex=sharex, sharey=sharey, squeeze=False)
            # Recursive call.
            for ax, index in zip(ax_list.ravel(), from_site_index):
                self.plot_site_pairs_total(index, what=what, exchange_xy=exchange_xy, ax=ax,
                    fontsize=fontsize, show=False)

            return fig

        # Single site
        pairs = [p for p in self.total if from_site_index == p[0]]
        if not pairs:
            cprint("Cannot find pairs starting from site index %s" % str(from_site_index), "yellow")
            return None

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)

        xlabel, ylabel, ysign = r"$E - E_f\;(eV)$", "COOP", +1
        if self.cop_type == "cohp":
            ylabel, ysign = "-COHP", -1

        # self.total[pair][spin][what]
        for pair in pairs:
            for spin in range(self.nsppol):
                xs, ys = self.energies, ysign * self.total[pair][spin][what]
                if exchange_xy: xs, ys = ys, xs
                label, style = self.get_labelstyle_from_spin_pair(spin, pair)
                ax.plot(xs, ys, label=label, **style)

        if exchange_xy:
            # Add vertical line to signal the zero.
            ax.axvline(c="k", ls=":", lw=1)
        else:
            ax.axhline(c="k", ls=":", lw=1)

        set_ax_xylabels(ax, xlabel, ylabel, exchange_xy)
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    def get_labelstyle_from_spin_pair(self, spin, pair):
        """Return label and linestyle for spin and pair indices"""
        type0, type1 = self.type_of_index[pair[0]], self.type_of_index[pair[1]]
        label = r"${%s}@{%s} \rightarrow {%s}@{%s}$" % (type0, pair[0], type1, pair[1])
        style = {"color": "black", "linewidth": 1.0} if spin == 0 else \
                {"color": "red", "linewidth": 1.0}
        # TODO: Improve style
        style = dict(lw=2, ls="-" if spin == 0 else "--") #, color=color)
        return label, style

    @add_fig_kwargs
    def plot_site_pairs_partial(self, from_site_index, what="single", exchange_xy=True, ax=None,
                                fontsize=8, **kwargs) -> Figure:
        """
        Plot partial crystal orbital projections (DOS or IDOS) for all sites listed in `from_site_index`

        Args:
            from_site_index: int of list of integers selecting the first site of the pairs to be included.
            what: "single" for COXP DOS, "integrated" for IDOS.
            exchange_xy: True to exchange x-y axis.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        if not self.partial:
            cprint("File does not contain partial projections", "yellow")
            return None

        # Handle list of sites.
        if duck.is_listlike(from_site_index):
            nrows, ncols = 1, len(from_site_index)
            sharex, sharey = True, False
            if exchange_xy: sharex, sharey = sharey, sharex
            ax_list, fig, plt = get_axarray_fig_plt(ax, nrows=nrows, ncols=ncols,
                                                    sharex=sharex, sharey=sharey, squeeze=False)
            # Recursive call.
            for ax, index in zip(ax_list.ravel(), from_site_index):
                self.plot_site_pairs_partial(index, what=what, exchange_xy=exchange_xy, ax=ax,
                    fontsize=fontsize, show=False)

            return fig

        # Single site.
        pairs = [p for p in self.partial if from_site_index == p[0]]
        if not pairs:
            cprint("Cannot find pairs starting from site index %s" % str(from_site_index), "yellow")
            return None

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)

        xlabel, ylabel, ysign = r"$E - E_f\;(eV)$", "pCOOP", +1
        if self.cop_type == "cohp":
            ylabel, ysign = "-COHP", -1

        # [(0, 1)]["4s", "4p_x"][spin]["single"]
        for pair in pairs:
            for orbs, d in self.partial[pair].items():
                for spin in range(self.nsppol):
                    xs, ys = self.energies, ysign * d[spin][what]
                    if exchange_xy: xs, ys = ys, xs
                    label, style = self.get_labelstyle_from_spin_pair_orbs(spin, pair, orbs)
                    ax.plot(xs, ys, label=label, **style)

        if exchange_xy:
            # Add vertical line to signal the zero.
            ax.axvline(c="k", ls=":", lw=2)
        else:
            ax.axhline(c="k", ls=":", lw=2)

        set_ax_xylabels(ax, xlabel, ylabel, exchange_xy)
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    def get_labelstyle_from_spin_pair_orbs(self, spin, pair, orbs):
        """Return label and linestyle for spin, pair indices and orbs tuple."""
        type0, type1 = self.type_of_index[pair[0]], self.type_of_index[pair[1]]
        label = r"$%s_{%s}@%s \rightarrow %s_{%s}@%s$" % (type0, orbs[0], pair[0], type1, orbs[1], pair[1])
        # TODO: Improve style
        #style = self.params_orbs[orbs[0]]["style"]
        style = dict(lw=2, ls="-" if spin == 0 else "--") #, color=color)
        return label, style

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("coxpfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(coxpfile)"),
            nbv.new_code_cell("coxpfile.plot(what='d');"),
            nbv.new_code_cell("coxpfile.plot(what='i');"),
            nbv.new_code_cell("coxpfile.plot_site_pairs_total(from_site_index=[0,]);"),
            nbv.new_code_cell("coxpfile.plot_site_pairs_partial(from_site_index=[0,]);"),
            nbv.new_code_cell("#coxpfile.plot_with_ebands(ebands='filepath');"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class ICoxpFile(_LobsterFile):
    """
    Wrapper class for the integrated crystal orbital projections up to the Fermi energy
    produced from Lobster.
    May contain the output stored in ICOHPLIST.lobster and ICOOPLIST.lobster

    .. attribute::  cop_type

        String. Either "coop" or "cohp".

    .. attribute:: values

        A dictionary with the following keys: a tuple with the index of the sites
        considered as a pair (0-based, e.g. (0,1)), the spin (i.e. 0 or 1)

    .. attribute:: type_of_index

        Dictionary mappping site index to element string.
    """

    @classmethod
    def from_file(cls, filepath):
        """
        Generates an instance of ICoxpFile from the files produce by Lobster.
        Accepts gzipped files.

        Args:
            filepath: path to the ICOHPLIST.lobster or ICOOPLIST.lobster.

        Returns:
            A ICoxpFile.
        """
        float_patt = r'-?(?:0|[1-9]\d*)(?:\.\d*)?(?:[eE][+\-]?\d+)?'
        header_patt = re.compile(r'.*?(over+\s#\s+bonds)?\s+for\s+spin\s+(\d).*')
        data_patt = re.compile(r'\s+\d+\s+([a-zA-Z]+)(\d+)\s+([a-zA-Z]+)(\d+)\s+(' +
                               float_patt + r')\s+(' + float_patt + r')(\d+)?')

        new = cls(filepath)
        new.values = tree()

        spin = None
        avg_num_bonds = False
        new.type_of_index = {}
        with zopen(filepath, "rt") as f:
            for line in f:
                match = header_patt.match(line.rstrip())
                if match:
                    spin = [0, 1][int(match.group(2))-1]
                    avg_num_bonds = match.group(1) is not None
                match = data_patt.match(line.rstrip())
                if match:
                    type1, index1, type2, index2, dist, avg, n_bonds = match.groups()
                    # 0-based indexing
                    index1 = int(index1) - 1
                    index2 = int(index2) - 1
                    new.type_of_index[index1] = type1
                    new.type_of_index[index2] = type2
                    avg_data = {'average': float(avg), 'distance': dist, 'n_bonds': int(n_bonds) if n_bonds else None}
                    new.values[(index1, index2)][spin] = avg_data
                    new.values[(index2, index1)][spin] = avg_data

        new.cop_type = "unknown"
        if "ICOOPLIST.lobster" in filepath: new.cop_type = "coop"
        if "ICOHPLIST.lobster" in filepath: new.cop_type = "cohp"

        return new

    def to_string(self, verbose=0):
        """String representation with verbosity level `verbose`."""
        lines = []; app = lines.append
        app("Number of pairs: %d" % len(self.values))
        app(self.dataframe.to_string(index=False))

        return "\n".join(lines)

    @lazy_property
    def dataframe(self):
        """|pandas-DataFrame| with results."""
        # self.values[pair][spin]
        import pandas as pd
        rows = []
        for pair, d in self.values.items():
            for spin in sorted(d.keys()):
                rows.append(OrderedDict([
                    ("index0", pair[0]),
                    ("index1", pair[1]),
                    ("type0", self.type_of_index[pair[0]]),
                    ("type1", self.type_of_index[pair[1]]),
                    ("spin", spin),
                    ("average", d[spin]["average"]),
                    ("distance", d[spin]["distance"]),
                    ("n_bonds", d[spin]["n_bonds"]),
                    ("pair", (pair[0], pair[1])),
                ]))

        return pd.DataFrame(rows, columns=list(rows[0].keys()))

    @add_fig_kwargs
    def plot(self, ax=None, **kwargs) -> Figure:
        """Barplot with average values."""
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        import seaborn as sns
        sns.barplot(x="average", y="pair", hue="spin", data=self.dataframe, ax=ax)
        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("icoxp_file = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(icoxp_file)"),
            nbv.new_code_cell("icoxp_file.plot();"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class LobsterDoscarFile(_LobsterFile):
    """
    Total and partial dos extracted from lobster DOSCAR.
    The fermi energy is always at the zero value.

    .. attribute:: energies

        List of energies. Shifted such that the Fermi level lies at 0 eV.

    .. attribute:: total_dos

        A dictionary with spin as a key (i.e. 0 or 1) and containing the values of the total DOS.
        Should have the same size as energies.

    .. attribute:: pdos

        A dictionary with the values of the projected DOS.
        The dictionary should have the following nested keys: the index of the site (0-based),
        the string representing the projected orbital (e.g. "4p_x"), the spin (i.e. 0 or 1).
        Each dictionary should contain a numpy array with a list of DOS values with the
        same size as energies.
    """

    @classmethod
    def from_file(cls, filepath):
        """
        Generates an instance from the DOSCAR.lobster file.
        Accepts gzipped files.

        Args:
            filepath: path to the DOSCAR.lobster.

        Returns:
            A LobsterDoscarFile.
        """
        with zopen(filepath, "rt") as f:
            dos_data = f.readlines()

        new = cls(filepath)

        new.nsites = int(dos_data[0].split()[0])
        n_energies = int(dos_data[5].split()[2])
        new.fermie = float(dos_data[5].split()[3])

        n_spin = 1 if len(dos_data[6].split()) == 3 else 2
        spins = [0, 1][:n_spin]

        # extract np array for total dos
        tdos_data = np.fromiter((d for l in dos_data[6:6+n_energies] for d in l.split()),
                                dtype=float).reshape((n_energies, 1+2*n_spin))

        new.energies = tdos_data[:,0].copy()
        new.total_dos = {}
        for i_spin, spin in enumerate(spins):
            new.total_dos[spin] = tdos_data[:,1+2*i_spin].copy()

        new.pdos = tree()
        new.type_of_index = {}
        # read partial doses
        for i_site in range(new.nsites):
            i_first_line = 5+(n_energies+1)*(i_site+1)

            # read orbitals
            # 6.01503759     -14.03508772   401       2.29842595       1.00000000; Z= 31; 4s 4p_y 4p_z 4p_x
            tokens = dos_data[i_first_line].split(';')
            orbitals = tokens[-1].split()
            Z = int(tokens[-2].split()[-1])
            el = Element.from_Z(Z)
            new.type_of_index[i_site] = el.symbol

            # extract np array for partial dos
            pdos_data = np.fromiter((d for l in dos_data[i_first_line+1:i_first_line+1+n_energies] for d in l.split()),
                dtype=float).reshape((n_energies, 1+n_spin*len(orbitals)))

            for i_orb, orb in enumerate(orbitals):
                for i_spin, spin in enumerate(spins):
                    new.pdos[i_site][orb][spin] = pdos_data[:, i_spin+n_spin*i_orb+1]

        new.nsppol = len(new.total_dos)
        return new

    def to_string(self, verbose=0):
        """String representation with Verbosity level `verbose`."""
        lines = []; app = lines.append
        app("Number of energies: %d, from %.3f to %.3f (eV) with E_fermi set to 0 (was %.3f)" % (
            len(self.energies), self.energies[0], self.energies[-1], self.fermie))
        app("nsppol: %d" % (self.nsppol))
        app("Number of sites in projected DOS: %d" % len(self.pdos))
        for i_site, dsite in self.pdos.items():
            app("%d --> {%s}" % (i_site, ", ".join(dsite.keys())))

        return "\n".join(lines)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)
        for site_index in range(self.nsites):
            yield self.plot_pdos_site(site_index, show=False)

    @add_fig_kwargs
    def plot(self, spin=None, ax=None, exchange_xy=False, fontsize=12, **kwargs) -> Figure:
        """
        Plot DOS.

        Args:
            spin:
            exchange_xy: True to exchange x-y axis.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)

        spins = range(self.nsppol) if spin is None else [spin]
        for spin in spins:
            opts = {"color": "black", "linewidth": 2.0, "ls": "-"} if spin == 0 else \
                   {"color": "red", "linewidth": 2.0, "ls": "--"}
            opts.update(kwargs)
            spin_sign = +1 if spin == 0 else -1
            xs, ys = self.energies, spin_sign * self.total_dos[spin]
            if exchange_xy: xs, ys = ys, xs
            ax.plot(xs, ys, label=self.spin2tex[spin] if self.nsppol == 2 else None,
                    **opts)

        xlabel, ylabel = r"$E - E_f\;(eV)$", "DOS"
        set_ax_xylabels(ax, xlabel, ylabel, exchange_xy)

        if self.nsppol == 2:
            ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_pdos_site(self, site_index, ax=None, exchange_xy=False, fontsize=8, **kwargs) -> Figure:
        """
        Plot projected DOS

        Args:
            site_index
            exchange_xy: True to exchange x-y axis.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        # Handle list of sites.
        if duck.is_listlike(site_index):
            nrows = len(site_index)
            sharex, sharey = True, False
            if exchange_xy: sharex, sharey = sharey, sharex
            ax_list, fig, plt = get_axarray_fig_plt(ax, nrows=nrows, ncols=1,
                                                    sharex=sharex, sharey=sharey, squeeze=False)
            # Recursive call.
            for ax, index in zip(ax_list.ravel(), site_index):
                self.plot_pdos_site(index, exchange_xy=exchange_xy, ax=ax, fontsize=fontsize, show=False)

            return fig

        # Single site.
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)

        # self.pdos[site_index]["4p_x"][spin]
        for orb, d in self.pdos[site_index].items():
            for spin in range(self.nsppol):
                spin_sign = +1 if spin == 0 else -1
                xs, ys = self.energies, spin_sign * d[spin]
                if exchange_xy: xs, ys = ys, xs
                n, lm = orb[0], orb[1:]
                params = self.params_orbs[lm]
                label = "$%s@%s_{%s}$" % (self.type_of_index[site_index], site_index, params["latex"])
                ax.plot(xs, ys, label=label, **params["style"])

        xlabel, ylabel = r"$E - E_f\;(eV)$", "PDOS"
        set_ax_xylabels(ax, xlabel, ylabel, exchange_xy)

        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("lobdos = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(lobdos)"),
            nbv.new_code_cell("lobdos.plot();"),
            nbv.new_code_cell("pdos.plot_pdos_site(site_index=[0,]);"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class LobsterInput(object):
    """
    This object stores the basic variables for a Lobster input and generates the lobsterin file.
    """

    accepted_basis_sets = {"bunge", "koga", "pbevaspfit2015"}

    available_advanced_options = {"basisRotation", "writeBasisFunctions", "onlyReadVasprun.xml", "noMemoryMappedFiles",
                                  "skipPAWOrthonormalityTest", "doNotIgnoreExcessiveBands", "doNotUseAbsoluteSpilling",
                                  "skipReOrthonormalization", "doNotOrthogonalizeBasis", "forceV1HMatrix",
                                  "noSymmetryCorrection", "symmetryDetectionPrecision", "useOriginalTetrahedronMethod",
                                  "useDecimalPlaces", "forceEnergyRange"}

    def __init__(self, basis_set=None, basis_functions=None, include_orbitals=None, atom_pairs=None, dist_range=None,
                 orbitalwise=True, start_en=None, end_en=None, en_steps=None, gaussian_smearing=None,
                 bwdf=None, advanced_options=None):
        """
        Args
            basis_set: String containing one of the possible basis sets available: bunge, koga, pbevaspfit2015
            basis_functions: list of strings giving the symbol of each atom and the basis functions: "Ga 4s 4p"
            include_orbitals: string containing which types of valence orbitals to use. E.g. "s p d"
            atom_pairs: list of tuples containing the couple of elements for which the COHP analysis will be
             performed. Index is 1-based.
            dist_range: list of tuples, each containing the minimum and maximum distance (in Angstrom) used to
             automatically generate atom pairs. Each tuple can also contain two atomic symbol to restric the match to
             the specified elements. examples: (0.5, 1.5) or (0.5, 1.5, 'Zn', 'O')
            start_en: starting energy with respect to the Fermi level (in eV)
            end_en: ending energy with respect to the Fermi level (in eV)
            en_steps: number of energy increments
            gaussian_smearing: smearing in eV when using gaussian broadening
            bwdf: enables the bond-weighted distribution function (BWDF). Value is the binning interval
            advanced_options: dict with additional advanced options. See lobster user guide for further details
        """
        if basis_set and basis_set.lower() not in self.accepted_basis_sets:
            raise ValueError("Wrong basis set {}".format(basis_set))
        self.basis_set = basis_set
        self.basis_functions = basis_functions or []
        self.include_orbitals = include_orbitals
        self.atom_pairs = atom_pairs or []
        self.dist_range = dist_range or []
        self.orbitalwise = orbitalwise
        self.start_en = start_en
        self.end_en = end_en
        self.en_steps = en_steps
        self.gaussian_smearing = gaussian_smearing
        self.bwdf = bwdf
        self.advanced_options = advanced_options or {}

        if not all(opt in self.available_advanced_options for opt in self.advanced_options.keys()):
            raise ValueError("Unknown adavanced options")

    @classmethod
    def _get_basis_functions_from_abinit_pseudos(cls, pseudos):
        """
        Extracts the basis function used from the PAW abinit pseudopotentials

        Args:
            pseudos: a list of Pseudos objects.
        """
        basis_functions = []
        for p in pseudos:
            if not hasattr(p, "valence_states"):
                raise RuntimeError("Only PAW pseudos in PAWXML format are supported by Lobster interface.")
            el = p.symbol + " ".join(str(vs['n'] + OrbitalType(int(vs['l'])).name)
                                     for vs in p.valence_states.values() if 'n' in vs)
            basis_functions.append(el)
        return basis_functions

    def set_basis_functions_from_abinit_pseudos(self, pseudos):
        """
        Sets the basis function used from the PAW abinit pseudopotentials

        Args:
            pseudos: a list of Pseudos objects.
        """
        basis_functions = self._get_basis_functions_from_abinit_pseudos(pseudos)

        self.basis_functions = basis_functions

    @classmethod
    def _get_basis_functions_from_potcar(cls, potcar):
        """
        Extracts the basis function used from a POTCAR.

        Args:
            potcar: a pymatgen.io.vasp.inputs.Potcar object
        """
        basis_functions = []
        for p in potcar:
            basis_functions.append(p.element + " " + " ".join(str(vs[0]) + vs[1] for vs in p.electron_configuration))
        return basis_functions

    def set_basis_functions_from_potcar(self, potcar):
        """
        Sets the basis function used from a POTCAR.

        Args:
            potcar: a pymatgen.io.vasp.inputs.Potcar object
        """
        basis_functions = self._get_basis_functions_from_potcar(potcar)

        self.basis_functions = basis_functions

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        lines = []

        if self.basis_set:
            lines.append("basisSet {}".format(self.basis_set))

        for bf in self.basis_functions:
            lines.append("basisFunctions {}".format(bf))

        for ap in self.atom_pairs:
            line = "cohpBetween atom {} atom {}".format(*ap)
            if self.orbitalwise:
                line += " orbitalwise"
            lines.append(line)

        for dr in self.dist_range:
            line = "cohpGenerator from {} to {}".format(dr[0], dr[1])
            if len(dr) > 2:
                line += " type {} type {}".format(dr[2], dr[3])
            if self.orbitalwise:
                line += " orbitalwise"
            lines.append(line)

        if self.start_en:
            lines.append("COHPStartEnergy {}".format(self.start_en))

        if self.end_en:
            lines.append("COHPEndEnergy {}".format(self.end_en))

        if self.en_steps:
            lines.append("COHPSteps {}".format(self.en_steps))

        if self.gaussian_smearing:
            lines.append("gaussianSmearingWidth {}".format(self.gaussian_smearing))

        if self.bwdf:
            lines.append("BWDF {}".format(self.bwdf))

        for k, v in self.advanced_options.items():
            lines.append("{} {}".format(k, v))

        return "\n".join(lines)

    @classmethod
    def from_dir(cls, dirpath, dE=0.01, **kwargs):
        """
        Generates an instance of the class based on the output folder of a DFT calculation.
        Reads the information from the pseudopotentials in order to determine the
        basis functions.

        Args:
            dirpath: the path to the calculation directory. For abinit it should contain the
                "files" file and GSR file, for vasp it should contain the vasprun.xml and the POTCAR.
            dE: The spacing of the energy sampling in eV.
            kwargs: the inputs for the init method, except for basis_functions, start_en,
                end_en and en_steps.

        Returns:
            A LobsterInput.
        """
        # These two libraries take a long time to import on HPC (07/02/24) so moved to method instead of header
        from pymatgen.io.vasp.outputs import Vasprun
        from pymatgen.io.vasp.inputs import Potcar

        # Try to determine the code used for the calculation
        dft_code = None
        if os.path.isfile(os.path.join(dirpath, 'vasprun.xml')):
            dft_code = "vasp"
            vr = Vasprun(os.path.join(dirpath, 'vasprun.xml'))

            en_min = np.min([bands_spin for bands_spin in vr.eigenvalues.values()])
            en_max = np.max([bands_spin for bands_spin in vr.eigenvalues.values()])
            fermie = vr.efermi

            potcar = Potcar.from_file(os.path.join(dirpath, 'POTCAR'))
            basis_functions = cls._get_basis_functions_from_potcar(potcar)

        elif glob.glob(os.path.join(dirpath, '*.files')):
            dft_code = "abinit"
            ff = glob.glob(os.path.join(dirpath, '*.files'))[0]
            with open(ff, "rt") as files_file:
                ff_lines = files_file.readlines()
            out_path = ff_lines[3].strip()
            if not os.path.isabs(out_path):
                out_path = os.path.join(dirpath, out_path)

            with GsrFile.from_file(out_path + '_GSR.nc') as gsr:
                en_min = gsr.ebands.eigens.min()
                en_max = gsr.ebands.eigens.max()
                fermie = gsr.ebands.fermie

            pseudo_paths = []
            for l in ff_lines[5:]:
                l = l.strip()
                if l:
                    if not os.path.isabs(l):
                        l = os.path.join(dirpath, l)
                    pseudo_paths.append(l)

            pseudos = [Pseudo.from_file(p) for p in pseudo_paths]

            basis_functions = cls._get_basis_functions_from_abinit_pseudos(pseudos)
        else:
            raise ValueError('Unable to determine the code used in dir {}'.format(dirpath))

        start_en = en_min + fermie
        end_en = en_max - fermie

        # shift the energies so that are divisible by dE and the value for the fermi level (0 eV) is included
        start_en = np.floor(start_en / dE) * dE
        end_en = np.ceil(end_en / dE) * dE

        en_steps = int((end_en - start_en) / dE)

        return cls(basis_functions=basis_functions, start_en=start_en, end_en=end_en, en_steps=en_steps, **kwargs)

    def write(self, dirpath='.'):
        """
        Write the input file 'lobsterin' in dirpath.
        """
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        # Write the input file.
        with open(os.path.join(dirpath, 'lobsterin'), "wt") as f:
            f.write(str(self))


class LobsterAnalyzer(NotebookWriter):

    @classmethod
    def from_dir(cls, dirpath, prefix=""):
        """
        Generates an instance of the class based on the output folder of a DFT calculation.

        Args:
            dirpath: the path to the calculation directory.
        """
        dirpath = os.path.abspath(dirpath)
        k2ext = {
            "coop_path": "COOPCAR.lobster",
            "cohp_path": "COHPCAR.lobster",
            "icohp_path": "ICOHPLIST.lobster",
            "lobdos_path": "DOSCAR.lobster",
        }
        kwargs = {k: None for k in k2ext}
        for k, ext in k2ext.items():
            # Handle [prefix]COHPCAR.lobster
            p = "%s*%s" % (prefix, ext)
            paths = glob.glob(os.path.join(dirpath, p))
            # Treat [prefix]COHPCAR.lobster.gz
            if not paths:
                p = "%s*%s.gz" % (prefix, ext)
                paths = glob.glob(os.path.join(dirpath, p))

            if paths:
                #print(k, "-->", paths)
                if len(paths) > 1:
                    raise RuntimeError("Found multiple files matching glob pattern: %s" % str(paths))
                kwargs[k] = paths[0]

        return cls(dirpath, prefix, **kwargs)

    def __init__(self, dirpath, prefix, coop_path=None, cohp_path=None, icohp_path=None, lobdos_path=None):
        self.coop = CoxpFile.from_file(coop_path) if coop_path else None
        self.cohp = CoxpFile.from_file(cohp_path) if cohp_path else None
        self.icohp = ICoxpFile.from_file(icohp_path) if icohp_path else None
        self.doscar = LobsterDoscarFile.from_file(lobdos_path) if lobdos_path else None
        self.dirpath = dirpath
        self.prefix = prefix

        for a in ("coop", "cohp", "icohp", "doscar"):
            obj = getattr(self, a)
            if obj is not None:
                self.nsppol = obj.nsppol
                break

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation with verbosity level `verbose`."""
        lines = []; app = lines.append
        for aname, header in [("coop", "COOP File"), ("cohp", "COHP File"),
                              ("icohp", "ICHOHPLIST File"), ("doscar", "Lobster DOSCAR"),]:
            obj = getattr(self, aname, None)
            if obj is None: continue
            app(marquee(header, mark="="))
            app(obj.to_string(verbose=verbose))
            app("")

        return "\n".join(lines)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)

    @add_fig_kwargs
    def plot(self, entries=("coop", "cohp", "doscar"), spin=None, **kwargs) -> Figure:
        """
        Plot COOP + COHP + DOSCAR.

        Args:
            entries: "cohp" to plot COHP, "coop" for COOP
            spin:

        Returns: |matplotlib-Figure|
        """
        entries = [e for e in entries if getattr(self, e, None)]
        if not entries: return None

        axmat, fig, plt = get_axarray_fig_plt(None, nrows=self.nsppol, ncols=len(entries),
                                              sharex=False, sharey=True, squeeze=False)

        spins = range(self.nsppol) if spin is None else [spin]
        for spin in spins:
            for ix, (label, ax) in enumerate(zip(entries, axmat[spin])):
                obj = getattr(self, label)
                obj.plot(ax=ax, spin=spin, exchange_xy=True, show=False)
                if ix != 0:
                    set_visible(ax, False, "ylabel")
                if self.nsppol == 2 and spin == 0:
                    set_visible(ax, False, "xlabel")

        return fig

    @add_fig_kwargs
    def plot_coxp_with_dos(self, from_site_index, what="cohp", with_orbitals=False, exchange_xy=True,
                           fontsize=8, **kwargs) -> Figure:
        """
        Plot COHP (COOP) for all sites in from_site_index and Lobster DOS on len(from_site_index) + 1 subfigures.

        Args:
            from_site_index: List of site indices for COHP (COOP)
            what: "cohp" to plot COHP, "coop" for COOP
            with_orbitals: True if orbital projections are wanted.
            exchange_xy: True to exchange x-y axis. By default, use x-axis for DOS-like quantities.
            fontsize: fontsize for legends and titles

        Returns: |matplotlib-Figure|
        """
        from_site_index = from_site_index if duck.is_listlike(from_site_index) else [from_site_index]
        ncols = len(from_site_index) + 1
        sharex, sharey = True, False
        if exchange_xy: sharex, sharey = sharey, sharex
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=1, ncols=ncols,
                                                sharex=sharex, sharey=sharey, squeeze=False)
        ax_list = ax_list.ravel()

        # Plot (COHP|COOP) (total|projections) on the first ncols - 1 axes.
        coxp_file = getattr(self, what)
        for ix, (ax, index) in enumerate(zip(ax_list[:-1], from_site_index)):
            if with_orbitals:
                coxp_file.plot_site_pairs_partial(index, exchange_xy=exchange_xy, ax=ax, fontsize=fontsize, show=False)
            else:
                coxp_file.plot_site_pairs_total(index, exchange_xy=exchange_xy, ax=ax, fontsize=fontsize, show=False)

            if ix != 0:
                set_visible(ax, False, "ylabel")
                #set_visible(ax, False, "xlabel")

        # Plot DOS on the last ax.
        ax = ax_list[-1]
        self.doscar.plot(ax=ax, exchange_xy=exchange_xy, fontsize=fontsize, show=False)
        set_visible(ax, False, "ylabel" if exchange_xy else "xlabel")

        return fig

    #@add_fig_kwargs
    #def plot_with_ebands(self, ebands, entries=("coop", "cohp", "doscar"), **kwargs) -> Figure:
    #    """
    #    Plot bands + COHP, COOP, DOSCAR.

    #    Args:
    #        ebands:
    #        entries: "cohp" to plot COHP, "coop" for COOP
    #        fontsize: fontsize for legends and titles

    #    Returns: |matplotlib-Figure|
    #    """
    #    ebands = ElectronBands.as_ebands(ebands)

    #    entries = [e for e in entries if getattr(self, e, None)]
    #    if not entries: return None

    #    import matplotlib.pyplot as plt
    #    from matplotlib.gridspec import GridSpec
    #    fig = plt.figure()

    #    # Build grid.
    #    nrows, ncols = self.nsppol, len(entries) + 1
    #    width_ratios = [2] + [1] * len(entries)
    #    gspec = GridSpec(nrows=nrows, ncols=ncols, width_ratios=width_ratios, wspace=0.05)

    #    # Bands and DOS will share the y-axis
    #    axmat = np.array((nrows, ncols), dtype=object)
    #    for icol in range(ncols):
    #        for irow in range(nrows):
    #            axmat[irow, icol] = plt.subplot(gspec[irow, icol], sharey=None if irow == 0 else axmat[irow, icol-1])

    #    #axmat, fig, plt = get_axarray_fig_plt(None, nrows=self.nsppol, ncols=len(entries) + 1,
    #    #                                        sharex=False, sharey=True, squeeze=False)

    #    for spin in range(self.nsppol):
    #        for ix, (label, ax) in enumerate(zip(entries, axmat[spin, :])):
    #            if ix == 0:
    #                ebands.plot(spin=spin, e0="fermie", ax=ax, show=False)
    #            else:
    #                obj = getattr(self, label)
    #                obj.plot(ax=ax, spin=spin, exchange_xy=True, show=False)

    #            if ix != 0:
    #                set_visible(ax, False, "ylabel")
    #            if self.nsppol == 2 and spin == 0:
    #                set_visible(ax, False, "xlabel")

    #    return fig

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("lobana = abilab.LobsterAnalyzer.from_dir(dirpath='%s', prefix='%s')" % (
                self.dirpath, self.prefix)),
            nbv.new_code_cell("print(lobana)"),
            nbv.new_code_cell("lobana.plot();"),
            nbv.new_code_cell("lobana.plot_coxp_with_dos(from_site_index=[0,], what='cohp', with_orbitals=False);"),
            #nbv.new_code_cell("lobana.plot_with_ebands(ebands='path_to_file_with_ebands');"),
        ])

        return self._write_nb_nbpath(nb, nbpath)



from phonopy import Phonopy
import numpy as np

from pymatgen.io.phonopy import get_pmg_structure,get_phonopy_structure

from abipy.core.abinit_units import eV_to_THz
from abipy.dfpt.converters import phonopy_to_abinit
import os,shutil
## Usefull functions to be used in embedding_ifc ###

def stru_0_1_to_minus_05_05(structure):
    """
    translate the structure so that the frac_coords go from [0,1] to [-0.5,0,5]
    """

    new_stru = structure.copy()
    for site in new_stru:
        if site.frac_coords[0] >= 0.5:
            site.frac_coords[0] = site.frac_coords[0] - 1

        if site.frac_coords[1] >= 0.5:
            site.frac_coords[1] = site.frac_coords[1] - 1

        if site.frac_coords[2] >= 0.5:
            site.frac_coords[2] = site.frac_coords[2] - 1
    return new_stru

def frac_coords_05_to_minus05(stru):
    """
    Move atoms that are close to frac_coord = 0.5 to -0.5
    Needed for atoms that are at -0.5 in the perfect supercell but move to for instance 0.498 after relaxation
    """
    new_stru = stru.copy()
    for i, atom in enumerate(new_stru):
        if atom.frac_coords[0] > 0.495:
            atom.frac_coords[0] = -0.5
        if atom.frac_coords[1] > 0.495:
            atom.frac_coords[1] = -0.5
        if atom.frac_coords[2] > 0.495:
            atom.frac_coords[2] = -0.5

    return new_stru

def center_wrt_defect(structure,defect_coord):
    """
    Center the structure around defect_coord. Defect is now at [0,0,0]
    """

    new_stru = structure.copy()
#    site = structure[index_in_structure]
    new_stru.translate_sites(indices=np.arange(0, len(structure)),
                            vector=-defect_coord,frac_coords=False,
                            to_unit_cell=False)
    return new_stru


def clean_structure(structure,defect_coord):
    """
    Apply successively :
    center_wrt_defect(), stru_0_1_to_minus_05_05(), frac_coords_05_to_minus05()
    Usefull to match structures after clean_structure()
    """
    stru=structure.copy()
    stru=center_wrt_defect(stru,defect_coord)
    stru=stru_0_1_to_minus_05_05(stru)
    stru=frac_coords_05_to_minus05(stru)
    return stru


def map_two_structures_coords(stru_1, stru_2,tol=0.1):
    """
    Returns a mapping between two structures based on 
    the coordinates, within a given tolerance in Angstrom.
    stru_1 should be a subset of the superset structure stru_2
    """

    cart_1 = stru_1.cart_coords
    cart_2 = stru_2.cart_coords
    mapping = []
    for i, site_1 in enumerate(stru_1):  # subset structure
        for j, site_2 in enumerate(stru_2):  # superset structure
            if max(abs(cart_1[i] - cart_2[j])) < tol:
                mapping.append(j)
    
    if len(mapping)!= len(stru_1):
        print(f"Mapping incomplete : only {len(mapping)}/{len(stru_1)} atoms mapped.")
    return mapping

def accoustic_sum(Hessian_matrix, atom_label):
    Sum_hessian=np.zeros((3,3))
    for m in range(len(Hessian_matrix[0])):
            Sum_hessian += Hessian_matrix[m][atom_label]
        
    return Sum_hessian


def inverse_participation_ratio(eigenvectors):
    """
    Compute equation (10) of https://pubs.acs.org/doi/10.1021/acs.chemmater.3c00537
    for a given q-point, eigenvectors shape=(nbands,natoms,3)
    """
    # for a given q-point,  eigenvectors shape=(nbands,natoms,3)
    ipr=[]
    for iband in range(len(eigenvectors)):    
        sum_atoms=0
        for iatom in range(len(eigenvectors[iband])):
            sum_atoms += np.dot(eigenvectors[iband,iatom],eigenvectors[iband,iatom])**2
        ipr.append(1/sum_atoms)
    return np.array(ipr).real

def localization_ratio(eigenvectors):
    """
    See equation (10) of https://pubs.acs.org/doi/10.1021/acs.chemmater.3c00537   
    for a given q-point, eigenvectors shape=(nbands,natoms,3)
    """
    ipr=inverse_participation_ratio(eigenvectors)
    return len(eigenvectors[0])/ipr


def vesta_phonon(eigenvectors,in_path,ibands=None,
                 scale_vector=20,width_vector=0.3,color_vector=[255,0,0],centered=True,
                 factor_keep_vectors=0.1,
                 out_path="VESTA_FILES"):
    """
    Draw the phonons eigenvectors on a vesta file. 
    Inspired from https://github.com/AdityaRoy-1996/Phonopy_VESTA/tree/master

    Args:
        eigenvectors: phonons eigenvectors for given q-point with shape (nbands,natom,3)
        in_path : path where the initial vesta structure in stored, the structure should be consistent 
        with the eigenvectors provided.
        ibands: list of indices of the bands to include, if not provided, all the bands
        scale_vector : scaling factor of the vector modulus
        width_vector : vector width
        color_vector : color in rgb format
        centered : center the vesta structure around [0,0,0]
        factor_keep_vectors : draw only the eigenvectors with magnitude > factor_keep_vectors * max(magnitude)
        out_path : path where .vesta files with vector are stored
    """

    vesta = open(in_path,'r').read()
    nbands = len(eigenvectors)
    natoms = len(eigenvectors[0])

    if ibands is None :
        ibands = range(nbands)

    if os.path.isdir(out_path): 
        shutil.rmtree(out_path)
        os.mkdir(out_path)
    else : 
        os.mkdir(out_path)

    path=out_path

    for iband in ibands : 
        towrite = vesta.split('VECTR')[0]
        towrite += 'VECTR\n'

        magnitudes=[]
        for iatom in range(natoms) :
            magnitudes.append(np.sqrt(eigenvectors[iband][iatom][0]**2+eigenvectors[iband][iatom][1]**2+eigenvectors[iband][iatom][2]**2))
        for iatom in range(natoms) :
            if magnitudes[iatom] > factor_keep_vectors * max(np.real(magnitudes)):
                towrite += '%5d' %(iatom + 1)
                towrite += '%10.5f' %(eigenvectors[iband][iatom][0] * int(scale_vector))
                towrite += '%10.5f' %(eigenvectors[iband][iatom][1] * int(scale_vector))
                towrite += '%10.5f' %(eigenvectors[iband][iatom][2] * int(scale_vector))
                towrite += '\n'
                towrite += '%5d' %(iatom + 1)  +  ' 0 0 0 0\n  0 0 0 0 0\n'
    
        towrite += '0 0 0 0 0\n' 
        towrite += 'VECTT\n'

        for atom in range(natoms) :
            towrite += '%5d' %(atom + 1)
            towrite += f'  {width_vector} {color_vector[0]}   {color_vector[1]}   {color_vector[2]} 0\n'

        towrite += '0 0 0 0 0\n' 
        towrite += 'SPLAN'
        towrite += vesta.split('SPLAN')[1]
        towrite += 'VECTS 1.00000'


        filename = path + f'/{iband:05}_'
        filename += '.vesta'

        open(filename, 'w').write(towrite)

        if centered==True:

            with open(filename, 'r') as file:
                file_contents = file.read()
                search_word="BOUND\n       0        1         0        1         0        1\n  0   0   0   0  0"
                replace_word="BOUND\n       -0.5        0.5         -0.5        0.5         -0.5        0.5\n  0   0   0   0  0"

                updated_contents = file_contents.replace(search_word, replace_word)

            with open(filename, 'w') as file:
                file.write(updated_contents)
            
    print(f"Vesta files created and stored in : \n {os.getcwd()}/{out_path}")
    return



#!/usr/bin/env python
r"""
Lobster COHP/COP/DOS
====================

This example shows how to analyze the output files
produced by Lobster code <<http://schmeling.ac.rwth-aachen.de/cohp/>

Use `abiview.py lobster DIRPATH` for a command line interface.
"""
import os
import abipy.data as abidata

from abipy.abilab import LobsterAnalyzer

dirpath = os.path.join(abidata.dirpath, "refs", "lobster_gaas")

# Open the all the lobster files produced in directory dirpath
# with the (optional) prefix GaAs_
lobana = LobsterAnalyzer.from_dir(dirpath, prefix="GaAs_")
print(lobana)

# Plot COOP + COHP + DOS.
lobana.plot(title="COOP + COHP + DOS")

# Plot COHP for all sites in from_site_index and Lobster DOS.
lobana.plot_coxp_with_dos(from_site_index=[0, 1])

# Plot orbital projections.
lobana.plot_coxp_with_dos(from_site_index=[0], with_orbitals=True)

#lobana.plot_with_ebands(ebands="out_GSR.nc")



#!/usr/bin/env python
"""
Wyckoff Flow
============

This example shows how to compute the band structure of a set of
crystalline structures obtained by changing the set of internal parameters (wyckoff positions).
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import os
import sys
import abipy.data as abidata

from abipy import abilab
from abipy import flowtk

def special_positions(lattice, u):
    """Construct the crystalline `Structure` for given value of the internal parameter u."""
    frac_coords = {}

    frac_coords["Si"] = [
            [1/2 + u, 1/2 - u, 0],
            [u,       -u,      u],
   ]
    frac_coords["O"] = [ [0, 0, u] ]

    species, coords = [], []
    for symbol, positions in frac_coords.items():
        species += len(positions) * [symbol]
        coords += positions

    return abilab.Structure(lattice, species, coords, validate_proximity=True, coords_are_cartesian=False)


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    pseudos = abidata.pseudos("14si.pspnc", "8o.pspnc")
    base_structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))

    news, uparams = [], [0.2, 0.3]
    for u in uparams:
        new = special_positions(base_structure.lattice, u)
        news.append(new)

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    # Create the list of workflows. Each workflow defines a band structure calculation.
    for new_structure, u in zip(news, uparams):
        # Generate the workflow and register it.
        flow.register_work(make_workflow(new_structure, pseudos))

    return flow


def make_workflow(structure, pseudos, paral_kgb=1):
    """
    Return a `Workflow` object defining a band structure calculation
    for given `Structure`.
    """

    # GS + NSCF run
    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=2)
    nval = structure.num_valence_electrons(pseudos)

    # Variables global to the SCF and the NSCF run.
    multi.set_vars(
        ecut=15,
        paral_kgb=paral_kgb,
        nband=nval//2 + 4,      # occupied + 4 empty
    )

    # (GS run)
    multi[0].set_kmesh(ngkpt=[4, 4, 4], shiftk=[0, 0, 0])
    multi[0].set_vars(tolvrs=1e-6)

    # (NSCF run)
    multi[1].set_vars(
        iscf=-2,
        tolwfr=1e-12,
    )
    multi[1].set_kpath(ndivsm=8)

    gs_inp, nscf_inp = multi.split_datasets()
    return flowtk.BandStructureWork(gs_inp, nscf_inp)


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).plot_networkx(tight_layout=True)


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
G0W0 convergence study
======================

G0W0 convergence study wrt ecuteps and the number of bands in W.
"""
import sys
import os
import numpy as np

import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk


def build_flow(options):

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Build the SCF input.
    scf_inp = abilab.AbinitInput(structure=abidata.structure_from_ucell("SiC"),
                                 pseudos=abidata.pseudos("14si.pspnc", "6c.pspnc"))

    ngkpt = [4, 4, 4]
    shiftk = [0, 0, 0]

    print("Number of valence electrons in the unit cell:", scf_inp.num_valence_electrons)

    scf_inp.set_vars(
        ecut=12,
        nband=10,
        ngkpt=ngkpt,
        shiftk=shiftk,
        tolvrs=1.e-4,
        #paral_kgb=1,
        #iomode=3,
    )

    #ebands_inp = scf_inp.make_ebands_input(ndivsm=15)

    # Build NSCF input on a k-mesh including empty states.
    nscf_inp = scf_inp.new_with_vars(
        nband=25,
        #nbdbuf=10
        tolwfr=1.e-8,
        iscf=-2
    )

    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Band structure work to produce the WFK file with a NSCF run + empty states.
    bands_work = flowtk.BandStructureWork(scf_inp, nscf_inp) # dos_inputs=[nscf_inp])
    flow.register_work(bands_work)

    # Compute screening by splitting all the q-points.
    ecuteps_list = np.arange(2, 8, 2)
    #ecuteps_list = [2]
    max_ecuteps = max(ecuteps_list)

    for ecuteps in ecuteps_list:
        scr_inp = nscf_inp.new_with_vars(
            optdriver=3,
            ecuteps=ecuteps,
            nband=20,
        )

        scr_work = flowtk.ScreeningWork.from_nscf_task(bands_work.nscf_task, scr_inp)

        # Ii you alredy have a WFK file and you want to skip the SCF + NSCF part
        # build the work using:

        #scr_work = ScreeningWork.from_wfk_file(wfk_filepath, scr_inp)

        scr_work.set_readme(f"SCR computation with ecuteps={ecuteps}")
        flow.register_work(scr_work)

    # Do a convergence study wrt ecuteps, each work is connected to a
    # different SCR file computed with a different value of nband.
    # Build a list of sigma inputs with different ecuteps

    sigma_work = flow.new_work()

    # Input file for SIGMA
    sigma_inp = scr_inp.new_with_vars(
        optdriver=4,
        nband=20,
        ecuteps=max_ecuteps,
        ecutsigx=scf_inp["ecut"],
        #ecutsigx=(3.5 * scf_input["ecut"]), ! This is problematic
        #gw_qprange
    )

    sigma_inp.set_kptgw(kptgw=[[0, 0, 0], [0.5, 0, 0]], bdgw=[1, 8])

    sigma_work.register_sigma_task(sigma_inp, deps={bands_work.nscf_task: "WFK", scr_work: "SCR"})

    #for scr_task in scr_work:
    #    sigma_conv = flowtk.SigmaConvWork(wfk_node=bands.nscf_task, scr_node=scr_task,
    #                                      sigma_inputs=sigma_inputs)
    #    flow.register_work(sigma_conv)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Ground-state SCF cycle
======================

This example shows how to plot the results of the
GS self-consistent cycle reported in the main output file.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the output file with GS calculation (Note the .abo extension).
# Alternatively, one can use `abiopen.py run.abo -nb`
# to generate a jupyter notebook.
abo = abiopen(abidata.ref_file("refs/si_ebands/run.abo"))

# Plot all SCF-GS sections found in the output file.
# Use abo.next_d2de_scf_cycle() for DFPT cycles.
scf_cycle = abo.next_gs_scf_cycle()
if scf_cycle is not None:
    scf_cycle.plot()

# If timopt -1, we can extract the timing and plot the data.
timer = abo.get_timer()
timer.plot_pie()

abo.close()



# coding: utf-8
"""Objects to run and analyze the frozen phonons generated from displacements of atoms"""
from __future__ import annotations

import numpy as np
import scipy.optimize as optimize

from monty.functools import lazy_property
from monty.collections import dict2namedtuple
from abipy.core.abinit_units import phfactor_ev2units, amu_emass, Bohr_Ang, eV_Ha
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt
from abipy.tools.typing import Figure


def quadratic_fit_function(xx, aa, bb):
    """
    Simple quadratic function used as a default for the fitting.

    Args:
        xx: the variable
        aa: the coefficient of the quadratic term
        bb: the constant term
    """
    return aa * xx ** 2 + bb


class FrozenPhonon:
    """
    Class defining a set of structures with displaced atoms.
    Provides methods to generate, interpolate and plot the data.
    """

    def __init__(self, original_structure, original_displ_cart, structures, normalized_displ_cart, etas,
                 qpt_frac_coords, scale_matrix, energies=None):
        """

        Args:
            original_structure: the original |Structure| object without any displacement.
            original_displ_cart: the original displacement used for generating the structures in cartesian
                coordinates. Should be linked to original_structure and should contain the phdispl_cart obtained
                from the phonon calculation.
            structures: list of structures with the displaced atoms
            normalized_displ_cart: displacement applied to the supercell to generate the list of structures.
                The normalization defines the largest displacement of one atom at 1 Angstrom.
            etas: list of (signed) factors that will multiply the normalized displacement to obtain the list of
                structure.
            qpt_frac_coords: fractional coordinates of the qpoint corresponding to the displacement.
            scale_matrix: the scaling matrix between the original cell and the supercell.
            energies: energies in eV corresponding to the structures.
        """

        self.original_structure = original_structure
        self.original_displ_cart = original_displ_cart
        self.structures = structures
        self.normalized_displ_cart = normalized_displ_cart
        self.etas = np.array(etas)
        self.qpt_frac_coords = qpt_frac_coords
        self.scale_matrix = scale_matrix
        self._energies = energies

    @property
    def energies(self):
        """
        Energies in eV corresponding to the the structures.
        """
        return self._energies

    @energies.setter
    def energies(self, energies):
        """
        Set the number of energies corresponding to the structures, should have the same length as structures
        """
        if len(energies) != self.n_displ:
            raise ValueError("The size of the enegies list does not match the number of internal structures.")
        self._energies = energies

    @property
    def qpt_cart_coords(self):
        """
        The cartesian coordinates of the qpoint.
        """
        return self.original_structure.lattice.reciprocal_lattice.get_cartesian_coords(self.qpt_frac_coords)

    @property
    def n_displ(self) -> int:
        """
        Number of displacements.
        """
        return len(self.structures)

    @lazy_property
    def ieta0(self) -> int:
        """
        The index corresponding to the structure with no displacements.
        """
        try:
            return self.etas.tolist().index(0)
        except ValueError:
            raise ValueError("The structure with no displacement is not present in the list.")

    @classmethod
    def from_phbands(cls, phbands, qpt_frac_coords, imode, etas, 
                     scale_matrix=None, max_supercell=None) ->FrozenPhonon:
        """
        Create an instace of FrozenPhonon using the eigendisplacements from a |PhononBands|

        Args:
            phbands: a |PhononBands| instance.
            qpt_frac_coords: q vector in reduced coordinate in reciprocal space or index of the qpoint.
            imode: index of the mode.
            etas: list of amplitudes of the displacement to be applied to the system. Will correspond to the
                largest displacement of one atom in Angstrom.
            scale_matrix: the scaling matrix of the supercell. If None a scaling matrix suitable for
                the qpoint will be determined.
            max_supercell: mandatory if scale_matrix is None, ignored otherwise. Defines the largest
                supercell in the search for a scaling matrix suitable for the q point.

        Returns:
            A FrozenPhonon.
        """

        qind = phbands.qindex(qpt_frac_coords)
        original_displ_cart = phbands.phdispl_cart[qind, imode].reshape((-1, 3))

        # first extract data with the normalized displacement (eta=1)
        normalized_fp = phbands.get_frozen_phonons(qind, imode, 1, scale_matrix, max_supercell)

        structures = []
        for n in etas:
            structures.append(phbands.get_frozen_phonons(qind, imode, n, normalized_fp.scale_matrix).structure)

        return cls(phbands.structure, original_displ_cart, structures, normalized_fp.displ, etas,
                   phbands.qpoints[qind].frac_coords, normalized_fp.scale_matrix)

    @lazy_property
    def mass_factor(self) -> float:
        """
        The factor accounting for the different masses and displacement of each atom
        """
        # the norms of each atomic displacement of the normalized displacement
        displ_norms = np.linalg.norm(self.normalized_displ_cart, axis=1)

        mass_factor = sum(site.specie.atomic_mass * (df**2)
                          for site, df in zip(self.structures[0], displ_norms / displ_norms.max())) / 2.
        mass_factor *= amu_emass
        return mass_factor

    def _freq_to_quad_coeff(self, freq):
        """
        Helper function to convert from a frequency in eV to the quadratic coefficient corresponding to the
        energies in eV and the etas in Angstrom. Uses the conversions to a.u. to get the correct value.
        """

        return freq ** 2 * self.mass_factor * eV_Ha / Bohr_Ang ** 2

    def _quad_coeff_to_freq(self, coeff):
        """
        Helper function to convert from the quadratic coefficient corresponding to the fit of energies in eV and
        the etas in Angstrom to a frequency in eV. Uses the conversions to a.u. to get the correct value.
        """

        return np.sqrt(coeff / self.mass_factor / eV_Ha) * Bohr_Ang

    def fit_to_frequency(self, fit_function=None, units="eV", min_fit_eta=None, max_fit_eta=None):
        """
        Uses the energies and the displacements to calculate the phonon frequency corresponding to the quadratic
        term of the fit.
        The fit is performed with scipy.optimize.curve_fit based on the function given in input and can also be
        limited number to a subset of the values of the displacements.

        Args:
            fit_function: a function that will be used to fit the data. The first parameter should be the coefficient
                of the quadratic term. If None a simple quadratic fit will be used.
            units: units of the output frequency. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz").
                Case-insensitive.
            min_fit_eta: if not None represents minimum value allowed for the (signed) eta to be used in the fit.
            max_fit_eta: if not None represents maximum value allowed for the (signed) eta to be used in the fit.

        Returns:
            A namedtuple with 'freq': the values of the frequency extracted from the fit,
            'fit_params': the parameters obtained from the fit, 'cov': the estimated covariance of fit_params
            (see scipy.optimize.curve_fit documentation for more details).
        """

        if self.energies is None:
            raise ValueError("The energies are required to calculate the fit")

        if min_fit_eta is None:
            min_fit_eta = self.etas.min()
        if max_fit_eta is None:
            max_fit_eta = self.etas.max()

        indices = np.where((min_fit_eta <= self.etas) & (self.etas <= max_fit_eta))

        if fit_function is None:
            fit_function = quadratic_fit_function

        etas = self.etas[indices]
        energies = np.array(self.energies)[indices]

        params, cov = optimize.curve_fit(fit_function, etas, energies)

        # frequency in eV.
        freq = self._quad_coeff_to_freq(params[0])

        return dict2namedtuple(freq=freq*phfactor_ev2units(units), fit_params=params, cov=cov)

    @add_fig_kwargs
    def plot_fit_energies(self, fit_function=None, min_fit_eta=None, max_fit_eta=None, freq=None,
                          ax=None, **kwargs) -> Figure:
        """
        Fits the displacements etas to the energies. See fit_to_frequency() for more details.

        Args:
            fit_function: a function that will be used to fit the data. The first parameter should be the coefficient
                of the quadratic term. If None a simple quadratic fit will be used.
            min_fit_eta: if not None represents minimum value allowed for the (signed) eta to be used in the fit.
            max_fit_eta: if not None represents maximum value allowed for the (signed) eta to be used in the fit.
            freq: if not None the quadratic function with this frequency as a coefficient will be added to the plot.
                freq in eV. Requires the 0 displacement to be present in the list of etas.
            ax: |matplotlib-Axes| or None if a new figure should be created.

        Returns:
            |matplotlib-Figure|
        """

        if fit_function is None:
            fit_function = quadratic_fit_function

        fit_data = self.fit_to_frequency(fit_function, min_fit_eta=min_fit_eta, max_fit_eta=max_fit_eta)

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        if "color" not in kwargs and "c" not in kwargs:
            kwargs["color"] = "blue"

        kwargs_points = dict(kwargs)
        kwargs_points["linestyle"] = "None"
        if "markersize" not in kwargs_points and "ms" not in kwargs_points:
            kwargs_points["markersize"] = 10
        if "marker" not in kwargs_points and "m" not in kwargs_points:
            kwargs_points["marker"] = "o"

        ax.plot(self.etas, self.energies, **kwargs_points)

        if "linewidth" not in kwargs and "lw" not in kwargs:
            kwargs["linewidth"] = 2
        kwargs["linestyle"] = "-"

        etas_plot = np.linspace(self.etas.min(), self.etas.max(), 100)
        ax.plot(etas_plot, fit_function(etas_plot, *fit_data.fit_params), **kwargs)

        if freq is not None:
            kwargs["linestyle"] = "--"
            e0 = self.energies[self.ieta0]
            en_freq = quadratic_fit_function(etas_plot, self._freq_to_quad_coeff(freq), e0)
            ax.plot(etas_plot, en_freq, **kwargs)

        ax.set_xlabel("Displacements (Ang)")
        ax.set_ylabel("Energy (eV)")

        return fig

    @add_fig_kwargs
    def plot_anharmonic_contribution(self, freq, relative=False, ax=None, **kwargs) -> Figure:
        """
        Plots the the absolute relative difference between the energies extracted from the frequency as
        quadratic coefficient and the calculated energies, giving an estimate of the anharmonic contribution
        Requires the 0 displacement to be present in the list of etas.

        Args:
            freq: phonon frequncy in eV
            relative: if True the plot will represent the relative difference with respect to the expected value
                obtained from the frequency, rather than the absolute difference.
            ax: |matplotlib-Axes| or None if a new figure should be created.

        Returns:
            |matplotlib-Figure|
        """

        if self.energies is None:
            raise ValueError("The energies are required to calculate the fit")

        self._freq_to_quad_coeff(freq)

        if "color" not in kwargs and "c" not in kwargs:
            kwargs["color"] = "blue"
        if "linewidth" not in kwargs and "lw" not in kwargs:
            kwargs["linewidth"] = 2

        e0 = self.energies[self.ieta0]
        en_freq = quadratic_fit_function(self.etas,self._freq_to_quad_coeff(freq), e0)

        diff = np.abs(en_freq - np.array(self.energies))

        if relative:
            diff = [d/(e-e0) * 100 if e != 0 else 0 for d, e in zip(diff, en_freq)]

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        ax.plot(self.etas, diff, **kwargs)

        ax.set_xlabel("Displacements (Ang)")
        if relative:
            ax.set_ylabel("Relative energy difference (%)")
        else:
            ax.set_ylabel("Energy difference (eV)")

        return fig



# coding: utf-8
"""Decorators for AbinitInput or MultiDataset objects."""
from __future__ import annotations
import abc
import pymatgen.io.abinit.abiobjects as aobj

from monty.json import MSONable
from abipy.tools.serialization import pmg_serialize
from abipy.flowtk.abiobjects import LdauParams, LexxParams
from .inputs import AbinitInput, MultiDataset


class InputDecoratorError(Exception):
    """Error class raised by :class:`AbinitInputDecorator`."""


class AbinitInputDecorator(MSONable, metaclass=abc.ABCMeta):
    """
    An `AbinitInputDecorator` adds new options to an existing :class:`AbinitInput`
    or an existing :class:`MultiDataset` without altering its structure. This is an abstract Base class.

    Example:

        decorator = MyDecorator(arguments)

        new_abinit_input = decorator(abinit_input)
        new_multidataset = decorator(multidataset)

    Note that a decorator does not modify the object on which it acts.

    .. warning::

        Please avoid introducing decorators acting on the structure (in particular the lattice)
        since the initial input may use the initial structure to  compute important variables.
        For instance, the list of k-points for band structure calculation depend on the bravais lattice
        and a decorator that changes it should recompute the path.
        This should  not represent a serious limitation because it's always possible to change the structure
        with its methods and then call the factory function without having to decorate an already existing object.
    """
    Error = InputDecoratorError

    def __str__(self):
        return str(self.as_dict())

    def __call__(self, obj, deepcopy=True):
        """
        Decorate an `AbinitInput` or a `MultiDataset` object
        This is the public API that calls the concrete implementation of the subclass

        Returns:
            New `AbinitInput` or new `MultiDataset` depending on obj.
        """

        if isinstance(obj, AbinitInput):
            new_inp = self._decorate(obj, deepcopy=deepcopy)
            # Log the decoration in new_inp.
            new_inp.register_decorator(self)
            return new_inp

        elif isinstance(obj, MultiDataset):
            new_inputs = []
            for inp in obj:
                new_inp = self._decorate(inp, deepcopy=deepcopy)
                # Log the decoration in new_inp.
                new_inp.register_decorator(self)
                new_inputs.append(new_inp)

            return MultiDataset.from_inputs(new_inputs)

        else:
            raise TypeError("Don't know how to decorate type %s" % type(obj))

    @abc.abstractmethod
    def _decorate(self, inp, deepcopy=True):
        """
        Abstract method that must be implemented by the concrete classes.
        It receives a :class:`AbinitInput` object, applies the decoration and returns a new `AbinitInput`.

        Args:
            inp: :class:`AbinitInput` object.
            deepcopy: True if a deepcopy of inp should be performed before changing the object.

        Returns:
            decorated :class:`AbinitInput` object (new object)
        """


class SpinDecorator(AbinitInputDecorator):
    """This decorator changes the spin polarization."""

    def __init__(self, spinmode, kptopt_ifspinor=4):
        self.spinmode = aobj.SpinMode.as_spinmode(spinmode)
        self.kptopt_ifspinor = kptopt_ifspinor

    @pmg_serialize
    def as_dict(self) -> dict:
        return dict(spinmode=self.spinmode.as_dict(), kptopt_ifspinor=self.kptopt_ifspinor)

    @classmethod
    def from_dict(cls, d: dict) -> SpinDecorator:
        return cls(aobj.SpinMode.from_dict(d["spinmode"]), kptopt_ifspinor=d["kptopt_ifspinor"])

    def _decorate(self, inp, deepcopy=True):
        if deepcopy: inp = inp.deepcopy()

        inp.set_vars(self.spinmode.to_abivars())

        # in version 7.11.5
        # When non-collinear magnetism is activated (nspden=4),
        # time-reversal symmetry cannot be used in the present
        # state of the code (to be checked and validated).
        # Action: choose kptopt different from 1 or 2.

        # Here we set kptopt to 4 (spatial symmetries, no time-reversal)
        # unless we already have a dataset with kptopt == 3 (no tr, no spatial)
        # This case is needed for DFPT at q != 0.
        if self.spinmode.nspinor == 2 and inp.get("kptopt") != 3:
            inp.set_vars(kptopt=self.kptopt_ifspinor)

        return inp


class SmearingDecorator(AbinitInputDecorator):
    """This decorator changes the electronic smearing."""

    def __init__(self, smearing):
        self.smearing = aobj.Smearing.as_smearing(smearing)

    @pmg_serialize
    def as_dict(self) -> dict:
        return {"smearing": self.smearing.as_dict()}

    @classmethod
    def from_dict(cls, d: dict) -> SmearingDecorator:
        return cls(aobj.Smearing.from_dict(d["smearing"]))

    def _decorate(self, inp, deepcopy=True):
        if deepcopy: inp = inp.deepcopy()
        inp.set_vars(self.smearing.to_abivars())
        return inp


class XcDecorator(AbinitInputDecorator):
    """Change the exchange-correlation functional."""

    def __init__(self, ixc: int):
        """
        Args:
            ixc: Abinit input variable
        """
        self.ixc = ixc

    @pmg_serialize
    def as_dict(self):
        return {"ixc": self.ixc}

    @classmethod
    def from_dict(cls, d: dict) -> XcDecorator:
        return cls(d["ixc"])

    def _decorate(self, inp, deepcopy=True):
        if deepcopy: inp = inp.deepcopy()
        # TODO: Don't understand why abinit does not enable usekden if MGGA!
        usekden = None
        #usekden = 1 if ixc.ismgga() else None
        inp.set_vars(ixc=self.ixc, usekden=usekden)

        return inp


class LdaUDecorator(AbinitInputDecorator):
    """This decorator adds LDA+U parameters to an :class:`AbinitInput` object."""

    def __init__(self, symbols_luj, usepawu=1, unit="eV"):
        """
        Args:
            symbols_luj: dictionary mapping chemical symbols to another dict with (l, u, j) values
            usepawu: Abinit input variable.
            unit: Energy unit for U and J
        """
        self.symbols_luj, self.usepawu, self.unit = symbols_luj, usepawu, unit

    @pmg_serialize
    def as_dict(self) -> dict:
        return dict(symbols_luj=self.symbols_luj, usepawu=self.usepawu, unit=self.unit)

    @classmethod
    def from_dict(cls, d: dict) -> LdaUDecorator:
        return cls(**{k: v for k, v in d.items() if not k.startswith("@")})

    def _decorate(self, inp, deepcopy=True):
        if not inp.ispaw: raise self.Error("LDA+U requires PAW!")
        if deepcopy: inp = inp.deepcopy()
        luj_params = LdauParams(usepawu=self.usepawu, structure=inp.structure)

        # Apply UJ on all the symbols present in symbols_lui.
        for symbol in inp.structure.symbol_set:
            if symbol not in self.symbols_luj: continue
            args = self.symbols_luj[symbol]
            luj_params.luj_for_symbol(symbol, l=args["l"], u=args["u"], j=args["j"], unit=self.unit)
            #luj_params.luj_for_symbol("Ni", l=2, u=u, j=0.1*u, unit=self.unit)

        inp.set_vars(luj_params.to_abivars())

        return inp


class LexxDecorator(AbinitInputDecorator):
    """This decorator add local exact exchange to a :class:`AbinitInput` object."""

    def __init__(self, symbols_lexx, exchmix=None):
        """
        Args:
            symbols_lexx: dictionary mapping chemical symbols to the angular momentum l on which lexx is applied.
            exchmix: ratio of exact exchange when useexexch is used. The default value of 0.25 corresponds to PBE0.

            Example. To perform a LEXX calculation for NiO in which the LEXX is computed only for the l=2
            channel of the nickel atoms:

                {"Ni": 2}
        """
        self.symbols_lexx, self.exchmix = symbols_lexx, exchmix

    @classmethod
    def from_dict(cls, d: dict) -> LexxDecorator:
        return cls(**{k:v for k, v in d.items() if not k.startswith("@")})

    @pmg_serialize
    def as_dict(self) -> dict:
        return {"symbols_lexx": self.symbols_lexx, "exchmix": self.exchmix}

    def _decorate(self, inp, deepcopy=True):
        if not inp.ispaw: raise self.Error("LEXX requires PAW!")
        if deepcopy: inp = inp.deepcopy()

        lexx_params = LexxParams(inp.structure)
        for symbol in inp.structure.symbol_set:
            if symbol not in self.symbols_lexx: continue
            lexx_params.lexx_for_symbol(symbol, l=self.symbols_lexx[symbol])

        # Context: the value of the variable useexexch is   1.
        # The value of the input variable ixc is    7, while it must be
        # equal to one of the following:  11  23
        # Action : you should change the input variables ixc or useexexch.
        inp.set_vars(lexx_params.to_abivars())
        dt_ixc = inp.get("ixc")
        if dt_ixc is None or dt_ixc not in [11, 23]: inp.set_vars(ixc=11)
        if self.exchmix is not None: inp.set_vars(exchmix=self.exchmix)

        return inp


# Stubs
#class ScfMixingDecorator(AbinitInputDecorator):


#class MagneticMomentDecorator(AbinitInputDecorator):
#    """Add reasoanble guesses for the initial magnetic moments."""


#class SpinOrbitDecorator(AbinitInputDecorator):
#    """Enable spin-orbit in the input."""
#     def __init__(self, no_spatial_symmetries=True, no_time_reversal=False, spnorbscl=None):
#        self.use_spatial_symmetries = use_spati
#        self.use_spatial_symmetries
#
#    def _decorate(self, inp, deepcopy=True)
#        if deepcopy: inp = inp.deepcopy()
#        kptopt =
#        if inp.ispaw:
#            for dt in inp.datasets:
#               dt.set_vars(pawspnorb=1, kptopt=kptopt)
#        return inp


#class PerformanceDecorator(AbinitInputDecorator):
#    """Change the variables in order to speedup the calculation."""
#    fftgw
#    boxcutmin
#    fft
#    def __init__(self, accuracy):
#        self.accuracy = accuracy
#
#    def _decorate(self, inp, deepcopy=True)
#        if deepcopy: inp = inp.deepcopy()
#         for dt in inp[1:]:
#            runlevel = dt.runlevel
#        return inp


#class DmftDecorator(AbinitInputDecorator):
#    """Add DMFT variables."""



#!/usr/bin/env python
r"""
Infrared spectrum of AlAs
=========================

This example shows how to plot the infrared spectrum of a polar semiconductor (AlAs)
from the DDB file  See tutorial/lesson_rf2.html

For a command line interfase, use:

    abiview.py ddb_ir in_DDB
"""
import os
import abipy.data as abidata

from abipy import abilab

# Open DDB file for alpha-SiO2 taken from https://materialsproject.org/materials/mp-7000/
filepath = os.path.join(abidata.dirpath, "refs", "mp-7000_DDB.bz2")
ddb = abilab.abiopen(filepath)

# Invoke anaddb to compute dielectric tensor and oscillator strength.
tgen = ddb.anaget_dielectric_tensor_generator(asr=2, chneut=1, dipdip=1, verbose=1)
print(tgen)

# Set phonon damping factor in eV (full width).
gamma_ev = 1e-3

# Plot IR spectrum in Cartesian coordinates.
tgen.plot_all(gamma_ev=gamma_ev, title="Diagonal and off-diagonal components")

tgen.plot(component="diag", reim="re", gamma_ev=gamma_ev, title="Real part, diagonal components")

tgen.plotly(component="diag", reim="re", gamma_ev=gamma_ev, title="Real part, diagonal components")

tgen.plot(component="diag", reim="im", gamma_ev=gamma_ev, title="Imaginary part, diagonal components")

tgen.plotly(component="diag", reim="im", gamma_ev=gamma_ev, title="Imaginary part, diagonal components")

ddb.close()



#!/usr/bin/env python
"""
Benchmark paral_kgb=1 algorithm with wfoptalg in [default, 1].
default corresponds to the LOBPCG algorithm, 1 enables the Chebyschev solver.
"""
import sys
import operator
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from functools import reduce
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input(paw=False):
    """
    Build and return an input file for GS calculations with paral_kgb=1
    """
    pseudos = abidata.pseudos("14si.pspnc", "8o.pspnc")
              #if not paw else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    structure = abidata.structure_from_ucell("SiO2-alpha")

    inp = abilab.AbinitInput(structure, pseudos)
    inp.set_kmesh(ngkpt=[4, 4, 4], shiftk=[0,0,0.5])

    # Global variables
    ecut = 40
    inp.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        nsppol=1,
        nband=32,
        paral_kgb=1,
        kptopt=3,
        istwfk="*1",
        timopt=-1,
        chksymbreak=0,
        nsym=1,
        prtwf=0,
        prtden=0,
        tolvrs=1e-8,
        nstep=20,
    )

    return inp


def build_flow(options):
    template = make_input()

    # Get the list of possible parallel configurations from abinit autoparal.
    #max_ncpus, min_eff = options.max_ncpus, options.min_eff
    #print("Getting all autoparal configurations up to max_ncpus: ",max_ncpus," with efficiency >= ",min_eff)
    #pconfs = template.abiget_autoparal_pconfs(max_ncpus, autoparal=1, verbose=options.verbose)
    #if options.verbose: print(pconfs)

    # Processor distribution.
    pconfs = [
       dict(npkpt=64, npband=1, npfft=2), # 128
       dict(npkpt=64, npband=2, npfft=2), # 256
       dict(npkpt=64, npband=2, npfft=4), # 512
       dict(npkpt=64, npband=4, npfft=4), # 1024
    ]

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    for wfoptalg in [None, 1]:
        work = flowtk.Work()
        for conf, omp_threads in product(pconfs, options.omp_list):
            #if not options.accept_conf(conf, omp_threads): continue
            mpi_procs = omp_threads * reduce(operator.mul, conf.values(), 1)

            manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
            inp = template.new_with_vars(conf, wfoptalg=wfoptalg)
            work.register_scf_task(inp, manager=manager)

        print("Found %d configurations" % len(work))
        flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
LDA+U calculations
==================

This example shows how to compute the LDA+U band structure of NiO
with PAW for several values of U-J.
"""

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk

from abipy.flowtk.abiobjects import LdauParams


def make_scf_nscf_dos_inputs(structure, pseudos, luj_params, paral_kgb=1):
    # Input file taken from tldau_2.in
    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=3)

    # Global variables
    global_vars = dict(
        #
        ecut=12,
        pawecutdg=30,
        nband=40,
        occopt=7,
        tsmear=0.015,
        nstep=50,
        paral_kgb=paral_kgb,
        #
        # Spin
        nsppol=1,
        nspden=2,
        nspinor=1,
        spinat=[0,  0,  1,
                0,  0, -1,
                0,  0,  0,
                0,  0,  0],
        # Kpoint Grid
        # The k point grid is not symmetric, but the calculations being
        # for the ground-state, this is not a problem.
    )

    multi.set_vars(global_vars)
    multi.set_vars(luj_params.to_abivars())

    # GS run.
    multi[0].set_vars(
        iscf=17,
        toldfe=1.0e-8,
        ngkpt=[2, 2, 2],
        chksymbreak=0,
    )

    # Band structure run.
    multi[1].set_kpath(ndivsm=6)
    multi[1].set_vars(tolwfr=1e-10)

    # DOS calculation.
    multi[2].set_vars(
        iscf=-3,   # NSCF calculation
        ngkpt=structure.calc_ngkpt(nksmall=8),
        shiftk=[0.0, 0.0, 0.0],
        nshiftk=1,
        tolwfr=1.e-8,
        #pawprtdos=1,
    )

    # Generate two input files for the GS and the NSCF run
    scf_input, nscf_input, dos_input = multi.split_datasets()

    return scf_input, nscf_input, dos_input


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    # Create the work for the band structure calculation.
    structure = abidata.structure_from_ucell("NiO")
    pseudos = abidata.pseudos("28ni.paw", "8o.2.paw")

    # The code below set up the parameters for the LDA+U calculation in NiO.
    #usepawu   1
    #lpawu   2 -1
    #upawu  8.0 0.0 eV
    #jpawu  0.8 0.0 eV
    usepawu = 1
    u_values = [5.0, 8.0]

    for u in u_values:
        # Apply U-J on Ni only.
        luj_params = LdauParams(usepawu, structure)
        luj_params.luj_for_symbol("Ni", l=2, u=u, j=0.1*u, unit="eV")

        scf_input, nscf_input, dos_input = make_scf_nscf_dos_inputs(structure, pseudos, luj_params)

        work = flowtk.BandStructureWork(scf_input, nscf_input, dos_inputs=dos_input)
        flow.register_work(work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
# Run the script with:
#
#     run_ldaus.py -s
#
# then use:
#
#    abirun.py flow_ldaus/ ebands
#
# to analyze all GSR files produced by flow.
#
# .. code-block:: bash
#
#        KS electronic bands:
#               nsppol  nspinor  nspden  nkpt  nband  nelect  fermie formula  natom  \
#        w0_t0       1        1       2     3     40    48.0   6.084  Ni2 O2      4
#        w0_t1       1        1       2   129     40    48.0   6.084  Ni2 O2      4
#        w0_t2       1        1       2   213     40    48.0   6.169  Ni2 O2      4
#        w1_t0       1        1       2     3     40    48.0   6.855  Ni2 O2      4
#        w1_t1       1        1       2   129     40    48.0   6.855  Ni2 O2      4
#        w1_t2       1        1       2   213     40    48.0   6.432  Ni2 O2      4
#
#               angle0  angle1  angle2      a      b      c  volume abispg_num  \
#        w0_t0    60.0    60.0    60.0  2.964  2.964  5.927  36.809        166
#        w0_t1    60.0    60.0    60.0  2.964  2.964  5.927  36.809        166
#        w0_t2    60.0    60.0    60.0  2.964  2.964  5.927  36.809        166
#        w1_t0    60.0    60.0    60.0  2.964  2.964  5.927  36.809        166
#        w1_t1    60.0    60.0    60.0  2.964  2.964  5.927  36.809        166
#        w1_t2    60.0    60.0    60.0  2.964  2.964  5.927  36.809        166
#
#                 scheme  occopt  tsmear_ev  bandwidth_spin0  fundgap_spin0  \
#        w0_t0  gaussian       7      0.408           99.343          3.324
#        w0_t1  gaussian       7      0.408           99.680          3.049
#        w0_t2  gaussian       7      0.408           99.838          2.560
#        w1_t0  gaussian       7      0.408           99.318          4.626
#        w1_t1  gaussian       7      0.408           99.628          3.352
#        w1_t2  gaussian       7      0.408           99.826          3.155
#
#               dirgap_spin0 task_class                               ncfile  node_id  \
#        w0_t0         3.351    ScfTask  flow_ldaus/w0/t0/outdata/out_GSR.nc   241313
#        w0_t1         3.352   NscfTask  flow_ldaus/w0/t1/outdata/out_GSR.nc   241314
#        w0_t2         3.041   NscfTask  flow_ldaus/w0/t2/outdata/out_GSR.nc   241315
#        w1_t0         4.626    ScfTask  flow_ldaus/w1/t0/outdata/out_GSR.nc   241317
#        w1_t1         3.928   NscfTask  flow_ldaus/w1/t1/outdata/out_GSR.nc   241318
#        w1_t2         3.983   NscfTask  flow_ldaus/w1/t2/outdata/out_GSR.nc   241319
#
#                           status
#        w0_t0  Completed
#        w0_t1  Completed
#        w0_t2  Completed
#        w1_t0  Completed
#        w1_t1  Completed
#        w1_t2  Completed
#
# The second task in each work (w*_t1) is a NSCF run along the high symmetry path.
# and we can compare the results of these two task by specifying the node identifiers:
#
#   abirun.py flow_ldaus/ ebands -p --nids=241314,241318
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_ldaus.png?raw=true
#    :alt: Band structure of Si in the IBZ and along a k-path
#
